"""Functions to parse dates and times."""

from __future__ import annotations

import re
from datetime import datetime, timedelta

_IVOA_TIMESTAMP_PATTERN = re.compile(
    r"(?P<timestamp>\d{4}-\d\d-\d\d(?P<time>T\d\d:\d\d:\d\d(\.\d\d\d)?)?)Z?$"
)
"""Regular expression matching an IVOA DALI timestamp."""

_TIMEDELTA_PATTERN = re.compile(
    r"((?P<weeks>\d+?)\s*(weeks|week|w))?\s*"
    r"((?P<days>\d+?)\s*(days|day|d))?\s*"
    r"((?P<hours>\d+?)\s*(hours|hour|hr|h))?\s*"
    r"((?P<minutes>\d+?)\s*(minutes|minute|mins|min|m))?\s*"
    r"((?P<seconds>\d+?)\s*(seconds|second|secs|sec|s))?$"
)
"""Regular expression pattern for a time duration."""

__all__ = [
    "parse_isodatetime",
    "parse_timedelta",
]


def parse_isodatetime(time_string: str) -> datetime:
    """Parse a string in a standard ISO date format.

    Parameters
    ----------
    time_string
        Date and time formatted as an ISO 8601 date and time, either using
        ``Z`` as the timezone or without timezone information. This is the
        same format produced by `isodatetime` and is compatible with
        Kubernetes and the IVOA DALI standard.

    Returns
    -------
    datetime.datetime
        The corresponding `~datetime.datetime`.

    Raises
    ------
    ValueError
        Raised if the provided time string is not in the correct format.

    Notes
    -----
    When parsing input for a model, use the `~safir.pydantic.IvoaIsoDatetime`
    type instead of this function. Using a model will be the normal case; this
    function is primarily useful in tests.
    """
    if m := re.match(_IVOA_TIMESTAMP_PATTERN, time_string):
        timestamp = m.group("timestamp")
        if not m.group("time"):
            timestamp += "T00:00:00"
        return datetime.fromisoformat(timestamp + "+00:00")
    else:
        raise ValueError(f"{time_string} does not match IVOA format")


def parse_timedelta(text: str) -> timedelta:
    """Parse a string into a `datetime.timedelta`.

    Expects a string consisting of one or more sequences of numbers and
    duration abbreviations, separated by optional whitespace. Whitespace at
    the beginning and end of the string is ignored. The supported
    abbreviations are:

    - Week: ``weeks``, ``week``, ``w``
    - Day: ``days``, ``day``, ``d``
    - Hour: ``hours``, ``hour``, ``hr``, ``h``
    - Minute: ``minutes``, ``minute``, ``mins``, ``min``, ``m``
    - Second: ``seconds``, ``second``, ``secs``, ``sec``, ``s``

    If several are present, they must be given in the above order. Example
    valid strings are ``8d`` (8 days), ``4h 3minutes`` (four hours and three
    minutes), and ``5w4d`` (five weeks and four days).

    If you want to accept strings of this type as input to a
    `~datetime.timedelta` field in a Pydantic model, use the
    `~safir.pydantic.HumanTimedelta` type as the field type. It uses this
    function to parse input strings.

    Parameters
    ----------
    text
        Input string.

    Returns
    -------
    datetime.timedelta
        Converted `datetime.timedelta`.

    Raises
    ------
    ValueError
        Raised if the string is not in a valid format.
    """
    m = _TIMEDELTA_PATTERN.match(text.strip())
    if m is None:
        raise ValueError(f"Could not parse {text!r} as a time duration")
    td_args = {k: int(v) for k, v in m.groupdict().items() if v is not None}
    return timedelta(**td_args)



"""Tests for the safir.logging module."""

from __future__ import annotations

import json
import logging
import re
from datetime import UTC, datetime, timedelta
from pathlib import Path
from unittest.mock import ANY

import pytest
import structlog
from httpx import AsyncClient
from pydantic import BaseModel, ValidationError

from safir.logging import LogLevel, Profile, configure_logging
from safir.testing.uvicorn import spawn_uvicorn

# Small app used to test Uvicorn logging configuration.
_UVICORN_APP = """
from fastapi import FastAPI
from safir.logging import configure_uvicorn_logging


app = FastAPI()
configure_uvicorn_logging()


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.get("/fail")
async def fail():
    raise TypeError("Uncaught exception")
"""


def _strip_color(string: str) -> str:
    """Strip ANSI color escape sequences.

    See https://stackoverflow.com/questions/14693701/
    """
    return re.sub(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]", "", string)


def test_log_level_enum() -> None:
    """Test that any case is allowed when initializing the enum."""
    assert LogLevel("warning") == LogLevel.WARNING
    assert LogLevel("WARNING") == LogLevel.WARNING
    assert LogLevel("Error") == LogLevel.ERROR

    # Check that this also works when going through Pydantic.
    class Model(BaseModel):
        log_level: LogLevel

    model = Model.model_validate({"log_level": "warning"})
    assert model.log_level == LogLevel.WARNING
    model = Model.model_validate({"log_level": "inFO"})
    assert model.log_level == LogLevel.INFO

    with pytest.raises(ValidationError):
        Model.model_validate({"log_level": "unknown"})


def test_configure_logging_development(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that development-mode logging is key-value formatted."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp", profile=Profile.development, log_level=LogLevel.INFO
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    app, level, line = caplog.record_tuples[0]
    assert app == "myapp"
    assert level == logging.INFO
    expected = "[info     ] Hello world                    [myapp] answer=42"
    assert _strip_color(line) == expected


def test_configure_logging_dev_timestamp(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test development-mode logging with an added timestamp."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp",
        profile="development",
        log_level="info",
        add_timestamp=True,
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.INFO
    match = re.match(
        (
            r"(\d+-\d+-\d+T\d+:\d+:[\d.]+Z) \[info\s+\] Hello world \s+"
            r" \[myapp\] answer=42"
        ),
        _strip_color(caplog.record_tuples[0][2]),
    )
    assert match
    isotimestamp = match.group(1)
    assert isotimestamp.endswith("Z")
    timestamp = datetime.fromisoformat(isotimestamp[:-1])
    timestamp = timestamp.replace(tzinfo=UTC)
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=5) < timestamp < now


def test_configure_logging_production(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that production-mode logging is JSON formatted."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp", profile=Profile.production, log_level=LogLevel.INFO
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert caplog.record_tuples[0] == (
        "myapp",
        logging.INFO,
        '{"answer": 42, "event": "Hello world", "logger": "myapp", '
        '"severity": "info"}',
    )


def test_configure_logging_prod_timestamp(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test development-mode logging with an added timestamp."""
    caplog.set_level(logging.INFO)

    configure_logging(
        name="myapp",
        profile="production",
        log_level="info",
        add_timestamp=True,
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.INFO
    data = json.loads(caplog.record_tuples[0][2])
    assert data == {
        "answer": 42,
        "event": "Hello world",
        "logger": "myapp",
        "severity": "info",
        "timestamp": ANY,
    }
    assert data["timestamp"].endswith("Z")
    timestamp = datetime.fromisoformat(data["timestamp"][:-1])
    timestamp = timestamp.replace(tzinfo=UTC)
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=5) < timestamp < now


def test_configure_logging_level(caplog: pytest.LogCaptureFixture) -> None:
    """Test that the logging level is set."""
    caplog.set_level(logging.DEBUG)

    configure_logging(name="myapp", log_level=LogLevel.INFO)
    logger = structlog.get_logger("myapp")

    logger.info("INFO message")
    assert len(caplog.record_tuples) == 1

    # debug-level shouldn't get logged
    logger.debug("DEBUG message")
    assert len(caplog.record_tuples) == 1


def test_duplicate_handlers(capsys: pytest.CaptureFixture[str]) -> None:
    """Test that configuring logging more than once doesn't duplicate logs."""
    configure_logging(name="myapp", profile="production", log_level="info")
    configure_logging(name="myapp", profile="production", log_level="info")

    logger = structlog.get_logger("myapp")

    logger.info("INFO not duplicate message")
    captured = capsys.readouterr()
    assert len(captured.out.splitlines()) == 1


def test_dev_exception_logging(caplog: pytest.LogCaptureFixture) -> None:
    """Test that exceptions are properly logged in the development logger."""
    configure_logging(
        name="myapp", profile=Profile.development, log_level=LogLevel.INFO
    )
    logger = structlog.get_logger("myapp")

    try:
        raise ValueError("this is some exception")
    except Exception:
        logger.exception("exception happened", foo="bar")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.ERROR
    assert "Traceback" in _strip_color(caplog.record_tuples[0][2])
    assert "this is some exception" in _strip_color(caplog.record_tuples[0][2])


def test_production_exception_logging(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that exceptions are properly logged in the production logger."""
    configure_logging(
        name="myapp", profile=Profile.production, log_level=LogLevel.INFO
    )
    logger = structlog.get_logger("myapp")

    try:
        raise ValueError("this is some exception")
    except Exception:
        logger.exception("exception happened", foo="bar")

    assert caplog.record_tuples[0][0] == "myapp"
    assert caplog.record_tuples[0][1] == logging.ERROR
    data = json.loads(caplog.record_tuples[0][2])
    assert data == {
        "event": "exception happened",
        "exception": ANY,
        "foo": "bar",
        "logger": "myapp",
        "severity": "error",
    }
    assert "Traceback (most recent call last)" in data["exception"]
    assert '"this is some exception"' in data["exception"]


@pytest.mark.asyncio
async def test_uvicorn_logging(tmp_path: Path) -> None:
    """Test the Uvicorn logging configuration."""
    app_path = tmp_path / "test.py"
    app_path.write_text(_UVICORN_APP)
    uvicorn = spawn_uvicorn(
        working_directory=tmp_path, app="test:app", capture=True
    )

    # Perform some operations against the app. Assume the pipe buffers are
    # large enough to hold all of the Uvicorn logging output generated by
    # this. (If they're ever not, we'll have to call communicate on the Popen
    # object.)
    try:
        async with AsyncClient() as client:
            r = await client.get(f"{uvicorn.url}/")
            assert r.status_code == 200
            r = await client.get(f"{uvicorn.url}/fail")
            assert r.status_code == 500
    finally:
        uvicorn.process.terminate()
        stdout, stderr = uvicorn.process.communicate()

    # stdout will be the access log. The raw event will contain the client
    # port, to which we do not have easy access, so only examine the
    # additional fields that should be added by Safir's custom Uvicorn log
    # handling.
    messages = [json.loads(line) for line in stdout.strip().split("\n")]
    assert messages == [
        {
            "event": ANY,
            "severity": "info",
            "httpRequest": {
                "protocol": "HTTP/1.1",
                "requestMethod": "GET",
                "requestUrl": "/",
                "status": "200",
            },
        },
        {
            "event": ANY,
            "severity": "info",
            "httpRequest": {
                "protocol": "HTTP/1.1",
                "requestMethod": "GET",
                "requestUrl": "/fail",
                "status": "500",
            },
        },
    ]

    # stderr will be the error log. There will be a bunch of random Uvicorn
    # messages here; just make sure they're all valid JSON. But make sure that
    # the log for the exception we raised includes the full traceback.
    #
    # Currently, FastAPI unconditionally emits a user warning with Python
    # 3.14, which we don't have formatting control over, so ignore those
    # lines.
    messages = [
        json.loads(line)
        for line in stderr.strip().split("\n")
        if "UserWarning" not in line and "pydantic.v1" not in line
    ]
    for message in messages:
        if "Exception in ASGI application" in message["event"]:
            assert "TypeError" in message["exception"]
            assert "in fail" in message["exception"]



"""Tests for asyncio utility functions."""

from __future__ import annotations

import asyncio

import pytest

from safir.asyncio import (
    AsyncMultiQueue,
    AsyncMultiQueueError,
    run_with_asyncio,
)


@pytest.mark.asyncio
async def test_async_multi_queue() -> None:
    queue = AsyncMultiQueue[str]()
    queue.put("one")
    queue.put("two")
    assert queue.qsize() == 2

    # Here and below, use this somewhat awkward construction to hide the
    # assertion from mypy's type narrowing of class attributes.
    finished = queue.finished
    assert not finished

    async def watcher(position: int) -> list[str]:
        return [s async for s in queue.aiter_from(position)]

    async def watcher_iter() -> list[str]:
        return [s async for s in queue]

    start_task = asyncio.create_task(watcher_iter())
    one_task = asyncio.create_task(watcher(1))
    current_task = asyncio.create_task(watcher(queue.qsize()))
    future_task = asyncio.create_task(watcher(3))
    way_future_task = asyncio.create_task(watcher(10))

    await asyncio.sleep(0.1)
    queue.put("three")
    await asyncio.sleep(0.1)
    queue.put("four")
    await asyncio.sleep(0.1)
    queue.clear()
    assert queue.qsize() == 0
    finished = queue.finished
    assert not finished

    after_clear_task = asyncio.create_task(watcher_iter())
    await asyncio.sleep(0.1)
    queue.put("something")
    queue.close()
    assert queue.finished
    assert queue.qsize() == 1

    assert await start_task == ["one", "two", "three", "four"]
    assert await one_task == ["two", "three", "four"]
    assert await current_task == ["three", "four"]
    assert await future_task == ["four"]
    assert await way_future_task == []
    assert await after_clear_task == ["something"]

    with pytest.raises(AsyncMultiQueueError):
        queue.put("else")


def test_run_with_asyncio() -> None:
    @run_with_asyncio
    async def some_async_function() -> str:
        return "some result"

    assert some_async_function() == "some result"



from ._exceptions import (
    FastStreamErrorHandler,
    IncompatibleSchemaError,
    InvalidAvroNameError,
    InvalidMetadataError,
    UnknownDeserializeError,
    UnknownSchemaError,
)
from ._kafka_config import (
    AIOKafkaParams,
    FastStreamBrokerParams,
    KafkaConnectionSettings,
    PlaintextSettings,
    SaslMechanism,
    SaslPlaintextSettings,
    SaslSslSettings,
    SecurityProtocol,
    SslSettings,
)
from ._manager import Compatibility, PydanticSchemaManager, SchemaInfo
from ._schema_registry_config import (
    SchemaManagerSettings,
    SchemaRegistryClientParams,
)

__all__ = [
    "AIOKafkaParams",
    "Compatibility",
    "FastStreamBrokerParams",
    "FastStreamErrorHandler",
    "IncompatibleSchemaError",
    "InvalidAvroNameError",
    "InvalidMetadataError",
    "KafkaConnectionSettings",
    "PlaintextSettings",
    "PydanticSchemaManager",
    "SaslMechanism",
    "SaslPlaintextSettings",
    "SaslSslSettings",
    "SchemaInfo",
    "SchemaManagerSettings",
    "SchemaRegistryClientParams",
    "SecurityProtocol",
    "SslSettings",
    "UnknownDeserializeError",
    "UnknownSchemaError",
]



"""Tests for Google Cloud Storage support code."""

from __future__ import annotations

from datetime import timedelta

import pytest

from safir.gcs import SignedURLService
from safir.testing.gcs import MockStorageClient


def test_signed_url(mock_gcs: MockStorageClient) -> None:
    url_service = SignedURLService("service-account", timedelta(hours=1))
    url = url_service.signed_url("s3://some-bucket/path/to/blob", "text/plain")
    assert url == "https://example.com/path/to/blob"

    # Test the same with a gs URL.
    url = url_service.signed_url("gs://some-bucket/path/to/blob", "text/plain")
    assert url == "https://example.com/path/to/blob"

    # Test that the lifetime is passed down to the mock, which will reject it
    # if it's not an hour.
    url_service = SignedURLService("foo", timedelta(minutes=30))
    with pytest.raises(AssertionError):
        url_service.signed_url("s3://some-bucket/blob", "text/plain")

    # Test that lifetime defaults to one hour.
    url_service = SignedURLService("foo")



"""Constants for metrics functionality."""

from __future__ import annotations

from datetime import timedelta

__all__ = [
    "ADMIN_CLIENT_PREFIX",
    "BROKER_PREFIX",
    "EVENT_MANAGER_DEFAULT_BACKOFF_INTERVAL",
    "EVENT_MANAGER_DEFAULT_KAFKA_TIMEOUT_MS",
]

ADMIN_CLIENT_PREFIX = "safir-metrics-admin-client"
"""Prefix for the client id for the metrics Kafka admin client."""

BROKER_PREFIX = "safir-metrics-faststream-broker"
"""Prefix for the Kafka client id for the metrics FastStream broker."""

EVENT_MANAGER_DEFAULT_BACKOFF_INTERVAL = timedelta(seconds=300)
"""Amount of time to wait before taking event manager out of error mode."""

EVENT_MANAGER_DEFAULT_KAFKA_TIMEOUT_MS = 1000
"""How long to wait for Kafka before raising an error for a blocking call."""



"""Tests for errors from the UWS API."""

from __future__ import annotations

from dataclasses import dataclass

import pytest
from httpx import AsyncClient

from safir.testing.sentry import Captured
from safir.testing.slack import MockSlackWebhook
from safir.testing.uws import MockWobbly
from safir.uws._dependencies import UWSFactory

from ..support.uws import SimpleParameters


@dataclass
class PostTest:
    """Encapsulates the data a test POST."""

    url: str
    data: dict[str, str]


@pytest.mark.asyncio
async def test_errors(
    client: AsyncClient,
    test_token: str,
    test_service: str,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
    mock_wobbly: MockWobbly,
    sentry_exception_items: Captured,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(
        test_token, SimpleParameters(name="June"), run_id="some-run-id"
    )

    # No token provided.
    routes = [
        "/test/jobs/1",
        "/test/jobs/1/destruction",
        "/test/jobs/1/error",
        "/test/jobs/1/executionduration",
        "/test/jobs/1/owner",
        "/test/jobs/1/parameters",
        "/test/jobs/1/phase",
        "/test/jobs/1/quote",
        "/test/jobs/1/results",
    ]
    for route in routes:
        request = client.build_request("GET", route)
        del request.headers["X-Auth-Request-Token"]
        r = await client.send(request)
        assert r.status_code == 422
        assert r.text.startswith("UsageError")

    # Wrong user specified.
    other_token = MockWobbly.make_token(test_service, "other-user")
    for route in routes:
        r = await client.get(
            route, headers={"X-Auth-Request-Token": other_token}
        )
        assert r.status_code == 404
        assert r.text.startswith("UsageError")

    # Job does not exist.
    for route in (r.replace("/1", "/2") for r in routes):
        r = await client.get(route)
        assert r.status_code == 404
        assert r.text.startswith("UsageError")

    # Check no user specified with POST routes.
    tests = [
        PostTest("/test/jobs/1", {"action": "DELETE"}),
        PostTest(
            "/test/jobs/1/destruction", {"destruction": "2021-09-10T10:01:02Z"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionduration": "1200"}
        ),
        PostTest("/test/jobs/1/phase", {"phase": "RUN"}),
    ]
    for test in tests:
        request = client.build_request("POST", test.url, data=test.data)
        del request.headers["X-Auth-Request-Token"]
        r = await client.send(request)
        assert r.status_code == 422
        assert r.text.startswith("UsageError")

    # Wrong user specified.
    for test in tests:
        r = await client.post(
            test.url,
            data=test.data,
            headers={"X-Auth-Request-Token": other_token},
        )
        assert r.status_code == 404
        assert r.text.startswith("UsageError")

    # Job does not exist.
    for test in tests:
        url = test.url.replace("/1", "/2")
        r = await client.post(url, data=test.data)
        assert r.status_code == 404
        assert r.text.startswith("UsageError")

    # Finally, test all the same things with the one supported DELETE.
    request = client.build_request("DELETE", "/test/jobs/1")
    del request.headers["X-Auth-Request-Token"]
    r = await client.send(request)
    assert r.status_code == 422
    assert r.text.startswith("UsageError")
    r = await client.delete(
        "/test/jobs/1", headers={"X-Auth-Request-Token": other_token}
    )
    assert r.status_code == 404
    assert r.text.startswith("UsageError")
    r = await client.delete("/test/jobs/2")
    assert r.status_code == 404
    assert r.text.startswith("UsageError")

    # Try some bogus destruction and execution duration parameters.
    tests = [
        PostTest("/test/jobs/1/destruction", {"destruction": "next tuesday"}),
        PostTest("/test/jobs/1/destruction", {"DESTruction": "next tuesday"}),
        PostTest(
            "/test/jobs/1/destruction", {"destruction": "2021-09-10 10:01:02"}
        ),
        PostTest(
            "/test/jobs/1/destruction",
            {"destrucTION": "2021-09-10T10:01:02+00:00:00"},
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionduration": "-1"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionDUration": "-1"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionduration": "fred"}
        ),
        PostTest(
            "/test/jobs/1/executionduration", {"executionDUration": "fred"}
        ),
    ]
    for test in tests:
        r = await client.post(test.url, data=test.data)
        assert r.status_code == 422, f"{test.url} {test.data}"
        assert r.text.startswith("UsageError"), r.text

    # Test bogus phase for async job creation.
    r = await client.post(
        "/test/jobs?phase=START",
        data={"runid": "some-run-id", "name": "Jane"},
    )
    assert r.status_code == 422
    r = await client.post(
        "/test/jobs",
        data={"runid": "some-run-id", "name": "Jane", "phase": "START"},
    )
    assert r.status_code == 422

    # None of these errors should have produced Slack or Sentry errors.
    assert mock_slack.messages == []
    assert sentry_exception_items.errors == []



"""Exceptions for the arq_ client."""

from __future__ import annotations

__all__ = [
    "ArqJobError",
    "JobNotFound",
    "JobNotQueued",
    "JobResultUnavailable",
]


class ArqJobError(Exception):
    """A base class for errors related to arq jobs."""

    def __init__(self, message: str, job_id: str | None) -> None:
        super().__init__(message)
        self._job_id = job_id

    @property
    def job_id(self) -> str | None:
        """The job ID, or `None` if the job ID is not known in this context."""
        return self._job_id


class JobNotQueued(ArqJobError):
    """The job was not successfully queued."""

    def __init__(self, job_id: str | None) -> None:
        super().__init__(
            f"Job was not queued because it already exists. id={job_id}",
            job_id,
        )


class JobNotFound(ArqJobError):
    """A job cannot be found."""

    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job could not be found. id={job_id}", job_id)


class JobResultUnavailable(ArqJobError):
    """The job's result is unavailable."""

    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job result could not be found. id={job_id}", job_id)



"""Signing Google Cloud Storage URLs."""

from __future__ import annotations

from datetime import timedelta
from urllib.parse import urlparse

try:
    import google.auth
    from google.auth import impersonated_credentials
    from google.cloud import storage
except ImportError as e:
    raise ImportError(
        "The safir.gcs module requires the gcs extra. "
        "Install it with `pip install safir[gcs]`."
    ) from e

__all__ = ["SignedURLService"]


class SignedURLService:
    """Generate signed URLs for Google Cloud Storage blobs.

    Uses default credentials plus credential impersonation to generate signed
    URLs for Google Cloud Storage blobs. This is the correct approach when
    running as a Kubernetes pod using workload identity.

    Parameters
    ----------
    service_account
        The service account to use to sign the URLs. The workload identity
        must have access to generate service account tokens for that service
        account.
    lifetime
        Lifetime of the generated signed URLs.

    Notes
    -----
    The workload identity (or other default credentials) under which the
    caller is running must have ``roles/iam.serviceAccountTokenCreator`` on
    the service account given in the ``service_account`` parameter. This is
    how a workload identity can retrieve a key that can be used to create a
    signed URL.

    See `gcs_signedurl <https://github.com/salrashid123/gcs_signedurl>`__ for
    additional details on how this works.
    """

    def __init__(
        self, service_account: str, lifetime: timedelta = timedelta(hours=1)
    ) -> None:
        self._lifetime = lifetime
        self._service_account = service_account
        self._gcs = storage.Client()
        self._credentials, _ = google.auth.default()

    def signed_url(self, uri: str, mime_type: str | None) -> str:
        """Generate signed URL for a given storage object.

        Parameters
        ----------
        uri
            URI for the storage object. This must start with ``s3://`` or
           ``gs://`` and use that URI syntax to specify bucket and blob of a
           Google Cloud Storage object.
        mime_type
            MIME type of the object, for encoding in the signed URL.

        Returns
        -------
        str
            New signed URL, which will be valid for as long as the lifetime
            parameter to the object.

        Raises
        ------
        ValueError
            Raised if the ``uri`` parameter is not an S3 or GCS URI.

        Notes
        -----
        This is inefficient, since it gets new signing credentials each time
        it generates a signed URL. Doing better will require figuring out the
        lifetime and refreshing the credentials when the lifetime has expired.
        """
        parsed_uri = urlparse(uri)
        if parsed_uri.scheme not in ("s3", "gs"):
            raise ValueError(f"URI {uri} is not an s3 or gs URI")
        bucket = self._gcs.bucket(parsed_uri.netloc)
        blob = bucket.blob(parsed_uri.path[1:])
        signing_credentials = impersonated_credentials.Credentials(
            source_credentials=self._credentials,
            target_principal=self._service_account,
            target_scopes=(
                "https://www.googleapis.com/auth/devstorage.read_only"
            ),
            lifetime=2,
        )
        return blob.generate_signed_url(
            version="v4",
            expiration=self._lifetime,
            method="GET",
            response_type=mime_type,
            credentials=signing_credentials,
        )



"""Kafka connection settings.

Adapted from Jonathan Sick's kafka work in `Ook <https://ook.lsst.io>`_.
"""

from __future__ import annotations

import ssl
from enum import StrEnum
from typing import Literal, NotRequired, Self, TypedDict

try:
    from aiokafka import helpers
    from faststream.security import (
        BaseSecurity,
        SASLPlaintext,
        SASLScram256,
        SASLScram512,
    )
except ImportError as e:
    raise ImportError(
        "The safir.kafka module requires the kafka extra. "
        "Install it with `pip install safir[kafka]`."
    ) from e
from pydantic import (
    AliasChoices,
    BaseModel,
    Field,
    FilePath,
    SecretStr,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict

__all__ = [
    "AIOKafkaParams",
    "FastStreamBrokerParams",
    "KafkaConnectionSettings",
    "PlaintextSettings",
    "PlaintextSettings",
    "SaslMechanism",
    "SaslPlaintextSettings",
    "SaslSslSettings",
    "SecurityProtocol",
    "SslSettings",
    "SslSettings",
]


class SecurityProtocol(StrEnum):
    """Kafka SASL security protocols."""

    SASL_PLAINTEXT = "SASL_PLAINTEXT"
    """Plain-text SASL-authenticated connection."""

    SASL_SSL = "SASL_SSL"
    """TLS-encrypted SASL-authenticated connection."""

    PLAINTEXT = "PLAINTEXT"
    """Plain-text connection."""

    SSL = "SSL"
    """SSL-encrypted SSL-authenticated connection."""


class SaslMechanism(StrEnum):
    """Kafka SASL mechanisms."""

    PLAIN = "PLAIN"
    """Plain-text SASL mechanism."""

    SCRAM_SHA_256 = "SCRAM-SHA-256"
    """SCRAM-SHA-256 SASL mechanism."""

    SCRAM_SHA_512 = "SCRAM-SHA-512"
    """SCRAM-SHA-512 SASL mechanism."""


class FastStreamBrokerParams(TypedDict):
    """Type for parameters to the constructor of a FastStream broker."""

    bootstrap_servers: str
    security: BaseSecurity


class AIOKafkaParams(TypedDict):
    """Type for parameters to the constructor of an aiokafka client."""

    bootstrap_servers: str
    security_protocol: str
    sasl_mechanism: NotRequired[str]
    sasl_plain_username: NotRequired[str]
    sasl_plain_password: NotRequired[str]
    ssl_context: NotRequired[ssl.SSLContext]


class SslSettings(BaseModel):
    """Subset of settings required for SSL auth."""

    bootstrap_servers: str

    security_protocol: Literal[SecurityProtocol.SSL]

    cluster_ca_path: FilePath

    client_cert_path: FilePath

    client_key_path: FilePath

    def to_ssl_context(self) -> ssl.SSLContext:
        """Create an SSL context for connecting to Kafka."""
        context = helpers.create_ssl_context(
            cafile=str(self.cluster_ca_path),
            certfile=str(self.client_cert_path),
            keyfile=str(self.client_key_path),
        )

        # Starting with Python 3.13, the VERIFY_X509_STRICT flag is set in
        # contexts created with ssl.create_default_context(), which this
        # context is. The certs generated by Strimzi on the Kafka bootstrap
        # endpoints do not have an Authority Key Identifier set, which makes
        # connections fail. Until these get fixed and regenerated in Strimzi,
        # unset VERIFY_X590_STRICT.
        context.verify_flags &= ~ssl.VERIFY_X509_STRICT
        return context

    def to_faststream_params(self) -> FastStreamBrokerParams:
        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security": BaseSecurity(ssl_context=self.to_ssl_context()),
        }

    def to_aiokafka_params(self) -> AIOKafkaParams:
        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security_protocol": self.security_protocol,
            "ssl_context": self.to_ssl_context(),
        }


class SaslPlaintextSettings(BaseModel):
    """Subset of settings required for SASL SSLauth."""

    bootstrap_servers: str

    security_protocol: Literal[
        SecurityProtocol.SASL_PLAINTEXT, SecurityProtocol.SASL_SSL
    ]

    sasl_mechanism: SaslMechanism = SaslMechanism.SCRAM_SHA_512

    sasl_username: str

    sasl_password: SecretStr

    def to_faststream_params(self) -> FastStreamBrokerParams:
        cls: type[SASLScram512 | SASLScram256 | SASLPlaintext]
        match self.sasl_mechanism:
            case SaslMechanism.SCRAM_SHA_512:
                cls = SASLScram512
            case SaslMechanism.SCRAM_SHA_256:
                cls = SASLScram256
            case SaslMechanism.PLAIN:
                cls = SASLPlaintext

        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security": cls(
                username=self.sasl_username,
                password=self.sasl_password.get_secret_value(),
            ),
        }

    def to_aiokafka_params(self) -> AIOKafkaParams:
        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security_protocol": self.security_protocol,
            "sasl_mechanism": self.sasl_mechanism,
            "sasl_plain_username": self.sasl_username,
            "sasl_plain_password": self.sasl_password.get_secret_value(),
        }


class SaslSslSettings(BaseModel):
    """Subset of settings required for SASL PLAINTEXT auth."""

    bootstrap_servers: str

    security_protocol: Literal[
        SecurityProtocol.SASL_PLAINTEXT, SecurityProtocol.SASL_SSL
    ]

    sasl_mechanism: SaslMechanism = SaslMechanism.SCRAM_SHA_512

    sasl_username: str

    sasl_password: SecretStr

    cluster_ca_path: FilePath | None

    def to_ssl_context(self) -> ssl.SSLContext:
        """Make an SSL context for connecting to Kafka."""
        cafile = None

        if self.cluster_ca_path:
            cafile = str(self.cluster_ca_path)
        context = helpers.create_ssl_context(
            cafile=cafile,
        )

        # Starting with Python 3.13, the VERIFY_X509_STRICT flag is set in
        # contexts created with ssl.create_default_context(), which this
        # context is. The certs generated by Strimzi on the Kafka bootstrap
        # endpoints do not have an Authority Key Identifier set, which makes
        # connections fail. Until these get fixed and regenerated in Strimzi,
        # unset VERIFY_X590_STRICT.
        context.verify_flags &= ~ssl.VERIFY_X509_STRICT
        return context

    def to_faststream_params(self) -> FastStreamBrokerParams:
        cls: type[SASLScram512 | SASLScram256 | SASLPlaintext]
        match self.sasl_mechanism:
            case SaslMechanism.SCRAM_SHA_512:
                cls = SASLScram512
            case SaslMechanism.SCRAM_SHA_256:
                cls = SASLScram256
            case SaslMechanism.PLAIN:
                cls = SASLPlaintext

        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security": cls(
                username=self.sasl_username,
                password=self.sasl_password.get_secret_value(),
                ssl_context=self.to_ssl_context(),
            ),
        }

    def to_aiokafka_params(self) -> AIOKafkaParams:
        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security_protocol": self.security_protocol,
            "ssl_context": self.to_ssl_context(),
            "sasl_mechanism": self.sasl_mechanism,
            "sasl_plain_username": self.sasl_username,
            "sasl_plain_password": self.sasl_password.get_secret_value(),
        }


class PlaintextSettings(BaseModel):
    """Subset of settings required for Plaintext auth."""

    bootstrap_servers: str

    security_protocol: Literal[SecurityProtocol.PLAINTEXT]

    def to_faststream_params(self) -> FastStreamBrokerParams:
        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security": BaseSecurity(),
        }

    def to_aiokafka_params(self) -> AIOKafkaParams:
        return {
            "bootstrap_servers": self.bootstrap_servers,
            "security_protocol": self.security_protocol,
        }


class KafkaConnectionSettings(BaseSettings):
    """Settings for connecting to Kafka.

    This settings model supports different authentication methods, which each
    have different sets of required settings. All of these settings can be
    provided in ``KAFKA_`` prefixed environment variables. Instances of this
    model have properties that can be used to construct different types of
    kafka clients:

    .. code-block:: python

       from faststream.broker import KafkaBroker

       from safir.kafka import KafkaConnectionSettings


       config = KafkaConnectionSettings()
       kafka_broker = KafkaBroker(**config.faststream_broker_params)

    When using this model directly, The ``validated`` property enforces at
    runtime that the correct settings were provided for the desired
    authentication method, and returns models to access those settings in a
    type-safe way:

    .. code-block:: python

       from pathlib import Path


       # ValidationError at runtime: ``client_key_path`` is not provided
       config = KafkaConnectionSettings(
           bootstrap_servers="something:1234",
           security_protocol=KafkaSecurityProtocol.SSL,
           cluster_ca_path=Path("/some/cert.crt"),
           client_cert_path=Path("/some/other/cert.crt"),
       )

       config = KafkaConnectionSettings(
           bootstrap_servers="something:1234",
           security_protocol=KafkaSecurityProtocol.SSL,
           cluster_ca_path=Path("/some/path/ca.crt"),
           client_cert_path=Path("/some/path/user.crt"),
           client_key_path=Path("/some/path/user.key"),
       )

       blah = config.validated.sasl_username  # Static type error
    """

    model_config = SettingsConfigDict(
        case_sensitive=False, extra="forbid", populate_by_name=True
    )

    bootstrap_servers: str = Field(
        title="Kafka bootstrap servers",
        description=(
            "A comma-separated list of Kafka brokers to connect to. "
            "This should be a list of hostnames or IP addresses, "
            "each optionally followed by a port number, separated by "
            "commas."
        ),
        examples=["kafka-1:9092,kafka-2:9092,kafka-3:9092", "kafka:9092"],
        validation_alias=AliasChoices(
            "bootstrapServers", "KAFKA_BOOTSTRAP_SERVERS"
        ),
    )

    security_protocol: SecurityProtocol = Field(
        title="Security Protocol",
        description=(
            "The authentication and encryption mode for the connection."
        ),
        validation_alias=AliasChoices(
            "securityProtocol", "KAFKA_SECURITY_PROTOCOL"
        ),
    )

    cluster_ca_path: FilePath | None = Field(
        default=None,
        title="Path to CA certificate file",
        description=(
            "The path to the PEM-formatted CA certificate file to use for "
            "verifying the broker's certificate. "
            "This is only needed for SSL and SASL_SSL security protocols, and"
            "even in those cases, only when the broker's certificate is not "
            "signed by a CA trusted by the operating system."
        ),
        examples=["/some/dir/ca.crt"],
        validation_alias=AliasChoices(
            "clusterCaPath", "KAFKA_CLUSTER_CA_PATH"
        ),
    )

    client_cert_path: FilePath | None = Field(
        default=None,
        title="Path to client certificate file",
        description=(
            "The path to the PEM-formated client certificate file to use for "
            "authentication. "
            "This is only needed if the broker is configured to require "
            "SSL client authentication."
        ),
        examples=["/some/dir/user.crt"],
        validation_alias=AliasChoices(
            "clientCertPath", "KAFKA_CLIENT_CERT_PATH"
        ),
    )

    client_key_path: FilePath | None = Field(
        default=None,
        title="Path to client key file",
        description=(
            "The path to the PEM-formatted client key file to use for "
            "authentication. This is only needed if for the SSL security"
            "protocol."
        ),
        examples=["/some/dir/user.key"],
        validation_alias=AliasChoices(
            "clientKeyPath", "KAFKA_CLIENT_KEY_PATH"
        ),
    )

    sasl_mechanism: SaslMechanism | None = Field(
        default=None,
        title="SASL mechanism",
        description=(
            "The SASL mechanism to use for authentication. "
            "This is only needed for the SASL_SSL and SASL_PLAINTEXT security"
            "protocols."
        ),
        validation_alias=AliasChoices("saslMechanism", "KAFKA_SASL_MECHANISM"),
    )

    sasl_username: str | None = Field(
        default=None,
        title="SASL username",
        description=(
            "The username to use for SASL authentication. "
            "This is only needed for the SASL_SSL and SASL_PLAINTEXT security"
            "protocols."
        ),
        validation_alias=AliasChoices("saslUsername", "KAFKA_SASL_USERNAME"),
    )

    sasl_password: SecretStr | None = Field(
        default=None,
        title="SASL password",
        description=(
            "The password to use for SASL authentication. "
            "This is only needed for the SASL_SSL and SASL_PLAINTEXT security"
            "protocols."
        ),
        validation_alias=AliasChoices("saslPassword", "KAFKA_SASL_PASSWORD"),
    )

    @model_validator(mode="after")
    def validate_auth_settings(self) -> Self:
        """Validate that the correct combination of parameters is specified."""
        _ = self.validated
        return self

    @property
    def validated(
        self,
    ) -> (
        SslSettings
        | SaslSslSettings
        | SaslPlaintextSettings
        | PlaintextSettings
    ):
        """Return a model with a subset of settings for a Kafka auth method.

        This method will fail with a ValidationError if an invalid set of
        settings were provided.
        """
        match self.security_protocol:
            case SecurityProtocol.PLAINTEXT:
                return PlaintextSettings(**self.model_dump())
            case SecurityProtocol.SASL_SSL:
                return SaslSslSettings(**self.model_dump())
            case SecurityProtocol.SASL_PLAINTEXT:
                return SaslPlaintextSettings(**self.model_dump())
            case SecurityProtocol.SSL:
                return SslSettings(**self.model_dump())

    def to_faststream_params(self) -> FastStreamBrokerParams:
        return self.validated.to_faststream_params()

    def to_aiokafka_params(self) -> AIOKafkaParams:
        return self.validated.to_aiokafka_params()



"""Helpers for testing Sentry reporting."""

from ._fixtures import capture_events_fixture, sentry_init_fixture
from ._mocks import Attachment, Captured, TestTransport

__all__ = [
    "Attachment",
    "Captured",
    "TestTransport",
    "capture_events_fixture",
    "sentry_init_fixture",
]



"""Tests for generic arq metrics functionality."""

from collections.abc import Callable
from typing import cast

import pytest
from arq.connections import ArqRedis, RedisSettings

from safir.metrics import (
    NOT_NONE,
    EventsConfiguration,
    MockEventPublisher,
    MockMetricsConfiguration,
)
from safir.metrics.arq import (
    ARQ_EVENTS_CONTEXT_KEY,
    ArqEvents,
    ArqMetricsError,
    initialize_arq_metrics,
    make_on_job_start,
    publish_queue_stats,
)


async def somejob(_ctx: dict) -> int:
    return 100


@pytest.mark.asyncio
async def test_arq_metrics(
    arq_redis: ArqRedis,
    create_arq_worker: Callable,
) -> None:
    # A variable to close over where we can put the container that
    # initialize_arq_metrics injects into the context so that we can assert
    # things about the events that were published
    events: ArqEvents | None = None

    # A startup function that initializes and event manager and calls
    # initialize_arq_metrics
    async def startup(ctx: dict) -> None:
        nonlocal events
        config = MockMetricsConfiguration(
            application="testapp",
            enabled=False,
            events=EventsConfiguration(topic_prefix="what.ever"),
            mock=True,
        )
        event_manager = config.make_manager()
        await event_manager.initialize()

        await initialize_arq_metrics(event_manager, ctx)
        events = ctx[ARQ_EVENTS_CONTEXT_KEY]

    # Create a real Arq worker
    queue_name = "some_queue"
    on_job_start = make_on_job_start(queue_name=queue_name)
    worker = create_arq_worker(
        functions=[somejob],
        queue_name=queue_name,
        on_startup=startup,
        on_job_start=on_job_start,
    )

    # Enqueue a job
    job = await arq_redis.enqueue_job("somejob", _queue_name=queue_name)
    assert job

    # Start the worker
    await worker.main()
    assert worker.jobs_complete == 1

    # Make sure the with_arq_metrics decorator didn't ruin our result
    result = await job.result()
    assert result == 100

    # Make sure the on_job_start function to published the right events
    assert events is not None
    publisher = cast("MockEventPublisher", events.arq_queue_job_event)
    publisher.published.assert_published_all(
        [
            {
                "time_in_queue": NOT_NONE,
                "queue": queue_name,
            }
        ]
    )


@pytest.mark.asyncio
async def test_arq_metrics_misconfiguration(
    arq_redis: ArqRedis,
    create_arq_worker: Callable,
) -> None:
    # A startup function that initializes and event manager but DOESN'T call
    # initialize_arq_metrics
    async def startup(ctx: dict) -> None:
        config = MockMetricsConfiguration(
            application="testapp",
            enabled=False,
            events=EventsConfiguration(topic_prefix="what.ever"),
            mock=True,
        )
        event_manager = config.make_manager()
        await event_manager.initialize()

    # Create a real Arq worker
    queue_name = "some_queue"
    on_job_start = make_on_job_start(queue_name=queue_name)
    worker = create_arq_worker(
        functions=[somejob],
        queue_name=queue_name,
        on_startup=startup,
        on_job_start=on_job_start,
    )
    # Enqueue a job
    job = await arq_redis.enqueue_job("somejob", _queue_name=queue_name)
    assert job

    # Start the worker
    with pytest.raises(ArqMetricsError):
        await worker.main()


@pytest.mark.asyncio
async def test_publish_queue_stats(arq_redis: ArqRedis) -> None:
    config = MockMetricsConfiguration(
        application="testapp",
        enabled=False,
        events=EventsConfiguration(topic_prefix="what.ever"),
        mock=True,
    )
    manager = config.make_manager()
    await manager.initialize()
    events = ArqEvents()
    await events.initialize(manager)

    queue_name = "some_queue"

    # Enqueue 5 jobs
    for _ in range(5):
        await arq_redis.enqueue_job("somejob", _queue_name=queue_name)

    # Publish stats
    host = arq_redis.connection_pool.connection_kwargs["host"]
    port = arq_redis.connection_pool.connection_kwargs["port"]

    redis_settings = RedisSettings(host=host, port=port)
    await publish_queue_stats(
        queue=queue_name, redis_settings=redis_settings, arq_events=events
    )

    # Check our published events
    publisher = cast("MockEventPublisher", events.arq_queue_stats)
    publisher.published.assert_published_all(
        [{"num_queued": 5, "queue": queue_name}]
    )



"""Tests for arq worker construction."""

from __future__ import annotations

import asyncio
import time
from datetime import timedelta
from typing import Any
from unittest.mock import ANY

import pytest
from arq.constants import default_queue_name
from arq.jobs import JobStatus
from arq.worker import Function
from structlog.stdlib import BoundLogger
from vo_models.uws.types import ErrorType, ExecutionPhase

from safir.arq import JobMetadata, MockArqQueue
from safir.arq.uws import (
    UWS_QUEUE_NAME,
    WorkerConfig,
    WorkerFatalError,
    WorkerJobInfo,
    WorkerResult,
    WorkerTimeoutError,
    build_worker,
)
from safir.datetime import current_datetime
from safir.testing.sentry import Captured
from safir.testing.slack import MockSlackWebhook
from safir.uws import JobResult, UWSApplication, UWSConfig
from safir.uws._constants import UWS_DATABASE_TIMEOUT
from safir.uws._dependencies import UWSFactory
from safir.uws._storage import JobStore

from ..support.uws import SimpleParameters


def hello(
    params: SimpleParameters, info: WorkerJobInfo, logger: BoundLogger
) -> list[WorkerResult]:
    if params.name == "Timeout":
        time.sleep(120)
    return [
        WorkerResult(
            result_id="greeting", url=f"https://example.com/{params.name}"
        )
    ]


@pytest.mark.asyncio
async def test_build_worker(
    uws_config: UWSConfig,
    test_token: str,
    test_username: str,
    logger: BoundLogger,
) -> None:
    redis_settings = uws_config.arq_redis_settings
    worker_config = WorkerConfig(
        arq_mode=uws_config.arq_mode,
        arq_queue_url=(
            f"redis://{redis_settings.host}:{redis_settings.port}"
            f"/{redis_settings.database}"
        ),
        arq_queue_password=redis_settings.password,
        grace_period=timedelta(seconds=60),
        parameters_class=SimpleParameters,
        timeout=uws_config.execution_duration,
    )
    settings = build_worker(hello, worker_config, logger)
    assert not settings.cron_jobs
    assert len(settings.functions) == 1
    assert isinstance(settings.functions[0], Function)
    assert settings.functions[0].name == hello.__qualname__
    assert settings.redis_settings == uws_config.arq_redis_settings
    assert settings.allow_abort_jobs
    assert settings.job_completion_wait == timedelta(seconds=60)
    assert settings.queue_name == default_queue_name
    assert settings.on_startup
    assert settings.on_shutdown

    # Run the startup hook.
    ctx: dict[Any, Any] = {}
    startup = settings.on_startup
    await startup(ctx)
    assert isinstance(ctx["arq"], MockArqQueue)
    assert isinstance(ctx["logger"], BoundLogger)
    arq = ctx["arq"]
    assert arq.default_queue_name == UWS_QUEUE_NAME

    # Run the worker.
    function = settings.functions[0].coroutine
    params = SimpleParameters(name="Roger")
    info = WorkerJobInfo(
        job_id="42",
        user=test_username,
        token=test_token,
        timeout=timedelta(minutes=1),
        run_id="some-run-id",
    )
    result = await function(ctx, params, info)
    assert result == [
        WorkerResult(result_id="greeting", url="https://example.com/Roger")
    ]
    assert list(arq._job_metadata[UWS_QUEUE_NAME].values()) == [
        JobMetadata(
            id=ANY,
            name="uws_job_started",
            args=(test_token, "42", ANY),
            kwargs={},
            enqueue_time=ANY,
            status=JobStatus.queued,
            queue_name=UWS_QUEUE_NAME,
        ),
        JobMetadata(
            id=ANY,
            name="uws_job_completed",
            args=(test_token, "42"),
            kwargs={},
            enqueue_time=ANY,
            status=JobStatus.queued,
            queue_name=UWS_QUEUE_NAME,
        ),
    ]

    # Run the shutdown hook.
    shutdown = settings.on_shutdown
    await shutdown(ctx)


@pytest.mark.asyncio
async def test_timeout(
    uws_config: UWSConfig,
    test_token: str,
    test_username: str,
    logger: BoundLogger,
) -> None:
    redis_settings = uws_config.arq_redis_settings
    worker_config = WorkerConfig(
        arq_mode=uws_config.arq_mode,
        arq_queue_url=(
            f"redis://{redis_settings.host}:{redis_settings.port}"
            f"/{redis_settings.database}"
        ),
        arq_queue_password=redis_settings.password,
        grace_period=timedelta(seconds=60),
        parameters_class=SimpleParameters,
        timeout=uws_config.execution_duration,
    )
    settings = build_worker(hello, worker_config, logger)
    assert isinstance(settings.functions[0], Function)

    # Run the startup hook.
    ctx: dict[Any, Any] = {}
    startup = settings.on_startup
    assert startup
    await startup(ctx)
    arq = ctx["arq"]

    # Run the worker.
    function = settings.functions[0].coroutine
    params = SimpleParameters(name="Timeout")
    info = WorkerJobInfo(
        job_id="42",
        user=test_username,
        token=test_token,
        timeout=timedelta(seconds=1),
        run_id="some-run-id",
    )
    with pytest.raises(WorkerTimeoutError):
        await function(ctx, params, info)
    assert list(arq._job_metadata[UWS_QUEUE_NAME].values()) == [
        JobMetadata(
            id=ANY,
            name="uws_job_started",
            args=(test_token, "42", ANY),
            kwargs={},
            enqueue_time=ANY,
            status=JobStatus.queued,
            queue_name=UWS_QUEUE_NAME,
        ),
        JobMetadata(
            id=ANY,
            name="uws_job_completed",
            args=(test_token, "42"),
            kwargs={},
            enqueue_time=ANY,
            status=JobStatus.queued,
            queue_name=UWS_QUEUE_NAME,
        ),
    ]

    # Make sure that handling the timeout didn't break the worker and we can
    # run another job successfully.
    params = SimpleParameters(name="Roger")
    info.job_id = "43"
    result = await function(ctx, params, info)
    assert result == [
        WorkerResult(result_id="greeting", url="https://example.com/Roger")
    ]

    # Run the shutdown hook.
    shutdown = settings.on_shutdown
    assert shutdown
    await shutdown(ctx)


@pytest.mark.asyncio
async def test_build_uws_worker(
    arq_queue: MockArqQueue,
    uws_config: UWSConfig,
    test_token: str,
    test_username: str,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
    sentry_exception_items: Captured,
    logger: BoundLogger,
) -> None:
    uws = UWSApplication(uws_config)
    job_service = uws_factory.create_job_service()
    job = await job_service.create(test_token, SimpleParameters(name="Ahmed"))
    results = [WorkerResult(result_id="greeting", url="https://example.com")]
    await job_service.start(test_token, test_username, job.id)
    job = await job_service.get(test_token, job.id)
    assert job.start_time is None
    assert job.phase == ExecutionPhase.QUEUED

    # Construct the arq configuration and check it.
    settings = uws.build_worker(logger)
    assert len(settings.functions) == 2
    job_started = settings.functions[0]
    assert callable(job_started)
    job_completed = settings.functions[1]
    assert callable(job_completed)
    assert settings.redis_settings == uws_config.arq_redis_settings
    assert not settings.allow_abort_jobs
    assert settings.job_completion_wait == UWS_DATABASE_TIMEOUT
    assert settings.queue_name == UWS_QUEUE_NAME
    assert settings.on_startup
    assert settings.on_shutdown

    # Run the startup hook.
    ctx: dict[Any, Any] = {}
    startup = settings.on_startup
    await startup(ctx)
    assert isinstance(ctx["arq"], MockArqQueue)
    assert isinstance(ctx["logger"], BoundLogger)
    assert isinstance(ctx["storage"], JobStore)
    assert ctx["arq"].default_queue_name == default_queue_name
    ctx["arq"] = arq_queue

    # Test starting a job.
    now = current_datetime()
    assert job.message_id
    await arq_queue.set_in_progress(job.message_id)
    await job_started(ctx, test_token, job.id, now)
    job = await job_service.get(test_token, job.id)
    assert job.phase == ExecutionPhase.EXECUTING
    assert job.start_time == now

    # Test finishing a job.
    assert job.message_id
    await asyncio.gather(
        job_completed(ctx, test_token, job.id),
        arq_queue.set_complete(job.message_id, result=results),
    )
    job = await job_service.get(test_token, job.id)
    assert job.phase == ExecutionPhase.COMPLETED
    assert job.end_time
    assert job.end_time.microsecond == 0
    assert now <= job.end_time <= current_datetime()
    assert job.results == [JobResult.from_worker_result(r) for r in results]
    assert mock_slack.messages == []
    assert sentry_exception_items.errors == []

    def nonnegative(value: int) -> None:
        if value < 0:
            raise ValueError("Value not nonnegative")

    def make_exception() -> None:
        try:
            nonnegative(-1)
        except Exception as e:
            raise WorkerFatalError(
                "Something", "went wrong", add_traceback=True
            ) from e

    # Test starting and erroring a job with a TaskError.
    job = await job_service.create(test_token, SimpleParameters(name="Ahmed"))
    await job_service.start(test_token, test_username, job.id)
    job = await job_service.get(test_token, job.id)
    assert job.message_id
    await arq_queue.set_in_progress(job.message_id)
    await job_started(ctx, test_token, job.id, now)
    try:
        make_exception()
    except WorkerFatalError as e:
        error = e
    await asyncio.gather(
        job_completed(ctx, test_token, job.id),
        arq_queue.set_complete(job.message_id, result=error, success=False),
    )
    job = await job_service.get(test_token, job.id)
    assert job.phase == ExecutionPhase.ERROR
    assert job.end_time
    assert job.end_time.microsecond == 0
    assert now <= job.end_time <= current_datetime()
    assert job.errors
    assert len(job.errors) == 1
    assert job.errors[0].type == ErrorType.FATAL
    assert job.errors[0].code == "Error"
    assert job.errors[0].message == "Something"
    assert job.errors[0].detail
    assert "went wrong" in job.errors[0].detail
    assert error.traceback
    assert error.traceback in job.errors[0].detail
    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "text": {
                        "text": "Error in vo-cutouts-db-worker: Something",
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
                {
                    "fields": [
                        {
                            "text": "*Exception type*\nTaskError",
                            "type": "mrkdwn",
                            "verbatim": True,
                        },
                        {
                            "text": "*UWS job ID*\n2",
                            "type": "mrkdwn",
                            "verbatim": True,
                        },
                        {"text": ANY, "type": "mrkdwn", "verbatim": True},
                        {"text": ANY, "type": "mrkdwn", "verbatim": True},
                        {
                            "text": f"*User*\n{test_username}",
                            "type": "mrkdwn",
                            "verbatim": True,
                        },
                    ],
                    "type": "section",
                },
                {
                    "text": {
                        "text": "*Original exception*\nValueError",
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
                {
                    "text": {
                        "text": "*Detail*\nwent wrong",
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
            ],
            "attachments": [
                {
                    "blocks": [
                        {
                            "text": {
                                "text": (
                                    f"*Traceback*\n```\n{error.traceback}```"
                                ),
                                "type": "mrkdwn",
                                "verbatim": True,
                            },
                            "type": "section",
                        }
                    ]
                }
            ],
        },
    ]

    traceback_attachment = sentry_exception_items.attachments[0]
    assert traceback_attachment.filename == "traceback"
    assert traceback_attachment.bytes.decode() == error.traceback

    sentry_error = sentry_exception_items.errors[0]
    assert sentry_error["tags"]["job_id"] == "2"
    assert sentry_error["contexts"]["info"]["started_at"] == ANY
    sentry_orig = sentry_error["contexts"]["original_exception"]
    assert sentry_orig["cause_type"] == "ValueError"
    assert sentry_orig["detail"] == "went wrong"

    sentry_exc = sentry_error["exception"]["values"][0]
    assert sentry_exc["type"] == "TaskError"
    assert sentry_exc["value"] == "Something"

    # Test starting and erroring a job with an unknown exception.
    mock_slack.messages = []
    job = await job_service.create(test_token, SimpleParameters(name="Ahmed"))
    await job_service.start(test_token, test_username, job.id)
    job = await job_service.get(test_token, job.id)
    assert job.message_id
    await arq_queue.set_in_progress(job.message_id)
    await job_started(ctx, test_token, job.id, now)
    exc = ValueError("some error")
    await asyncio.gather(
        job_completed(ctx, test_token, job.id),
        arq_queue.set_complete(job.message_id, result=exc, success=False),
    )
    job = await job_service.get(test_token, job.id)
    assert job.phase == ExecutionPhase.ERROR
    assert job.errors
    assert len(job.errors) == 1
    assert job.errors[0].type == ErrorType.FATAL
    assert job.errors[0].code == "Error"
    assert job.errors[0].message == "Unknown error executing task"
    assert job.errors[0].detail == "ValueError: some error"
    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "text": {
                        "text": "Error in vo-cutouts-db-worker",
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
                {
                    "fields": [
                        {
                            "text": "*Exception type*\nValueError",
                            "type": "mrkdwn",
                            "verbatim": True,
                        },
                        {"text": ANY, "type": "mrkdwn", "verbatim": True},
                    ],
                    "type": "section",
                },
                {
                    "text": {
                        "text": (
                            "*Exception*\n```\nValueError: some error\n```"
                        ),
                        "type": "mrkdwn",
                        "verbatim": True,
                    },
                    "type": "section",
                },
                {"type": "divider"},
            ],
        },
    ]

    sentry_exc = sentry_exception_items.errors[1]["exception"]["values"][0]
    assert sentry_exc["type"] == "ValueError"
    assert sentry_exc["value"] == "some error"



"""Tests for Pydantic utility functions."""

from __future__ import annotations

import json
from datetime import UTC, datetime, timedelta, timezone

import pytest
from pydantic import (
    BaseModel,
    ValidationError,
    field_validator,
    model_validator,
)

from safir.pydantic import (
    CamelCaseModel,
    EnvAsyncPostgresDsn,
    EnvRedisDsn,
    HumanTimedelta,
    IvoaIsoDatetime,
    SecondsTimedelta,
    UtcDatetime,
    normalize_datetime,
    normalize_isodatetime,
    to_camel_case,
    validate_exactly_one_of,
)


def test_env_async_postgres_dsn(monkeypatch: pytest.MonkeyPatch) -> None:
    class TestModel(BaseModel):
        dsn: EnvAsyncPostgresDsn

    monkeypatch.delenv("POSTGRES_5432_TCP_PORT", raising=False)
    monkeypatch.delenv("POSTGRES_HOST", raising=False)
    model = TestModel.model_validate(
        {"dsn": "postgresql://localhost:7777/some-database"}
    )
    assert model.dsn.scheme == "postgresql"
    assert not model.dsn.username
    assert not model.dsn.password
    assert model.dsn.host == "localhost"
    assert model.dsn.port == 7777
    assert model.dsn.path == "/some-database"
    assert not model.dsn.query

    model = TestModel.model_validate(
        {
            "dsn": (
                "postgresql+asyncpg://user:password@localhost/other"
                "?connect_timeout=10"
            )
        }
    )
    assert model.dsn.scheme == "postgresql+asyncpg"
    assert model.dsn.username == "user"
    assert model.dsn.password == "password"
    assert model.dsn.host == "localhost"
    assert not model.dsn.port
    assert model.dsn.path == "/other"
    assert model.dsn.query == "connect_timeout=10"

    monkeypatch.setenv("POSTGRES_5432_TCP_PORT", "8999")
    model = TestModel.model_validate(
        {
            "dsn": (
                "postgresql://user:password@localhost/other?connect_timeout=10"
            )
        }
    )
    assert model.dsn.scheme == "postgresql"
    assert model.dsn.username == "user"
    assert model.dsn.password == "password"
    assert model.dsn.host == "localhost"
    assert model.dsn.port == 8999
    assert model.dsn.path == "/other"
    assert model.dsn.query == "connect_timeout=10"

    monkeypatch.setenv("POSTGRES_HOST", "example.com")
    model = TestModel.model_validate({"dsn": "postgresql://localhost/other"})
    assert model.dsn.scheme == "postgresql"
    assert not model.dsn.username
    assert not model.dsn.password
    assert model.dsn.host == "example.com"
    assert model.dsn.port == 8999
    assert model.dsn.path == "/other"
    assert not model.dsn.query

    with pytest.raises(ValidationError):
        TestModel.model_validate(
            {"dsn": "postgresql+psycopg2://localhost/other"}
        )


def test_env_redis_dsn(monkeypatch: pytest.MonkeyPatch) -> None:
    class TestModel(BaseModel):
        dsn: EnvRedisDsn

    monkeypatch.delenv("REDIS_6379_TCP_PORT", raising=False)
    monkeypatch.delenv("REDIS_HOST", raising=False)
    model = TestModel.model_validate(
        {"dsn": "redis://user:password@example.com:7777/1"}
    )
    assert model.dsn.scheme == "redis"
    assert model.dsn.username == "user"
    assert model.dsn.password == "password"
    assert model.dsn.host == "example.com"
    assert model.dsn.port == 7777
    assert model.dsn.path == "/1"

    model = TestModel.model_validate({"dsn": "redis://localhost"})
    assert model.dsn.scheme == "redis"
    assert not model.dsn.username
    assert not model.dsn.password
    assert model.dsn.host == "localhost"
    assert model.dsn.port == 6379
    assert model.dsn.path == "/0"

    monkeypatch.setenv("REDIS_6379_TCP_PORT", "4567")
    model = TestModel.model_validate(
        {"dsn": "redis://user:password@example.com:7777/1"}
    )
    assert model.dsn.scheme == "redis"
    assert model.dsn.username == "user"
    assert model.dsn.password == "password"
    assert model.dsn.host == "example.com"
    assert model.dsn.port == 4567
    assert model.dsn.path == "/1"

    monkeypatch.setenv("REDIS_HOST", "127.12.0.1")
    model = TestModel.model_validate({"dsn": "redis://localhost"})
    assert model.dsn.scheme == "redis"
    assert not model.dsn.username
    assert not model.dsn.password
    assert model.dsn.host == "127.12.0.1"
    assert model.dsn.port == 4567
    assert model.dsn.path == "/0"

    with pytest.raises(ValidationError):
        TestModel.model_validate({"dsn": "rediss://example.com/0"})


def test_human_timedelta() -> None:
    class TestModel(BaseModel):
        delta: HumanTimedelta

    model = TestModel.model_validate({"delta": timedelta(seconds=5)})
    assert model.delta == timedelta(seconds=5)
    assert model.model_dump(mode="python") == {"delta": timedelta(seconds=5)}
    assert model.model_dump(mode="json") == {"delta": 5}
    model = TestModel.model_validate({"delta": "4h5m18s"})
    assert model.delta == timedelta(hours=4, minutes=5, seconds=18)
    model = TestModel.model_validate({"delta": 600})
    assert model.delta == timedelta(seconds=600)
    assert model.model_dump(mode="python") == {"delta": timedelta(seconds=600)}
    model = TestModel.model_validate({"delta": 4.5})
    assert model.delta.total_seconds() == 4.5
    assert model.model_dump(mode="json") == {"delta": 4.5}
    model = TestModel.model_validate({"delta": "300"})
    assert model.delta == timedelta(seconds=300)
    model = TestModel.model_validate({"delta": "10.5"})
    assert model.delta.total_seconds() == 10.5

    with pytest.raises(ValidationError):
        TestModel.model_validate({"delta": "P1DT12H"})


def test_seconds_timedelta() -> None:
    class TestModel(BaseModel):
        delta: SecondsTimedelta

    model = TestModel.model_validate({"delta": timedelta(seconds=5)})
    assert model.delta == timedelta(seconds=5)
    assert model.model_dump(mode="python") == {"delta": timedelta(seconds=5)}
    assert model.model_dump(mode="json") == {"delta": 5}
    model = TestModel.model_validate({"delta": 600})
    assert model.delta == timedelta(seconds=600)
    assert model.model_dump(mode="python") == {"delta": timedelta(seconds=600)}
    model = TestModel.model_validate({"delta": 4.5})
    assert model.delta.total_seconds() == 4.5
    assert model.model_dump(mode="json") == {"delta": 4.5}
    model = TestModel.model_validate({"delta": "300"})
    assert model.delta == timedelta(seconds=300)
    model = TestModel.model_validate({"delta": "10.5"})
    assert model.delta.total_seconds() == 10.5

    with pytest.raises(ValidationError):
        TestModel.model_validate({"delta": "P1DT12H"})


@pytest.mark.filterwarnings("ignore:.*datetime.utcnow.*:DeprecationWarning")
def test_utcdatetime() -> None:
    class TestModel(BaseModel):
        time: UtcDatetime

    date = datetime.fromtimestamp(1668814932, tz=UTC)
    model = TestModel.model_validate({"time": 1668814932})
    assert model.time == date
    assert model.model_dump(mode="python") == {"time": date}
    round_trip = datetime.fromisoformat(model.model_dump(mode="json")["time"])
    assert round_trip == date

    mst_zone = timezone(-timedelta(hours=7))
    mst_date = datetime.now(tz=mst_zone)
    utc_date = mst_date.astimezone(UTC)
    assert TestModel.model_validate({"time": mst_date}).time == utc_date

    naive_date = datetime.utcnow()  # noqa: DTZ003
    aware_date = TestModel.model_validate({"time": naive_date}).time
    assert aware_date == naive_date.replace(tzinfo=UTC)
    assert aware_date.tzinfo == UTC

    model = TestModel.model_validate({"time": "2023-01-25T15:44:00+00:00"})
    assert model.time == datetime.fromisoformat("2023-01-25T15:44:00+00:00")
    assert model.time.tzinfo == UTC


def test_ivoaisodatetime() -> None:
    class TestModel(BaseModel):
        time: IvoaIsoDatetime

    date = datetime.fromisoformat("2023-01-25T15:44:34+00:00")
    model = TestModel.model_validate({"time": "2023-01-25T15:44:34Z"})
    assert model.time == date
    assert model.model_dump(mode="python") == {"time": date}
    assert model.model_dump(mode="json") == {"time": "2023-01-25T15:44:34Z"}
    model = TestModel.model_validate({"time": "2023-01-25T15:44:34"})
    assert model.time == date

    date = datetime.fromisoformat("2023-01-25").replace(tzinfo=UTC)
    assert TestModel.model_validate({"time": "2023-01-25"}).time == date

    with pytest.raises(ValidationError, match="does not match IVOA format"):
        TestModel.model_validate({"time": "2023-01-25T15:44:00+00:00"})

    with pytest.raises(ValidationError, match="Must be a string in"):
        TestModel.model_validate({"time": 1668814932})

    with pytest.raises(ValidationError, match="does not match IVOA format"):
        TestModel.model_validate({"time": "next thursday"})


@pytest.mark.filterwarnings("ignore:.*datetime.utcnow.*:DeprecationWarning")
def test_normalize_datetime() -> None:
    class TestModel(BaseModel):
        time: datetime | None

        _val = field_validator("time", mode="before")(normalize_datetime)

    assert TestModel(time=None).time is None

    date = datetime.fromtimestamp(1668814932, tz=UTC)
    model = TestModel.model_validate({"time": 1668814932})
    assert model.time == date

    mst_zone = timezone(-timedelta(hours=7))
    mst_date = datetime.now(tz=mst_zone)
    utc_date = mst_date.astimezone(UTC)
    assert TestModel(time=mst_date).time == utc_date

    naive_date = datetime.utcnow()  # noqa: DTZ003
    aware_date = TestModel(time=naive_date).time
    assert aware_date == naive_date.replace(tzinfo=UTC)
    assert aware_date.tzinfo == UTC

    with pytest.raises(ValueError, match=r"Must be a datetime or seconds .*"):
        TestModel.model_validate({"time": "2023-01-25T15:44:00+00:00"})


def test_normalize_isodatetime() -> None:
    class TestModel(BaseModel):
        time: datetime | None

        _val = field_validator("time", mode="before")(normalize_isodatetime)

    assert TestModel(time=None).time is None

    date = datetime.fromisoformat("2023-01-25T15:44:34+00:00")
    model = TestModel.model_validate({"time": "2023-01-25T15:44:34Z"})
    assert model.time == date

    date = datetime.fromisoformat("2023-01-25T15:44:00+00:00")
    model = TestModel.model_validate({"time": "2023-01-25T15:44:00"})
    assert model.time == date

    date = datetime.fromisoformat("2023-01-25").replace(tzinfo=UTC)
    assert TestModel.model_validate({"time": "2023-01-25"}).time == date

    with pytest.raises(ValidationError, match="does not match IVOA format"):
        TestModel.model_validate({"time": "2023-01-25T15:44:00+00:00"})

    with pytest.raises(ValidationError, match="Must be a string in"):
        TestModel.model_validate({"time": 1668814932})

    with pytest.raises(ValidationError, match="does not match IVOA format"):
        TestModel.model_validate({"time": "1668814932"})


def test_to_camel_case() -> None:
    assert to_camel_case("foo") == "foo"
    assert to_camel_case("minimum_lifetime") == "minimumLifetime"
    assert to_camel_case("replace_403") == "replace403"
    assert to_camel_case("foo_bar_baz") == "fooBarBaz"


def test_camel_case_model() -> None:
    class TestModel(CamelCaseModel):
        minimum_lifetime: int
        replace_403: bool
        foo_bar_baz: str

    camel = {
        "minimumLifetime": 10,
        "replace403": False,
        "fooBarBaz": "something",
    }
    snake = {
        "minimum_lifetime": 10,
        "replace_403": False,
        "foo_bar_baz": "something",
    }
    data = TestModel.model_validate(camel)
    assert data.minimum_lifetime == 10
    assert not data.replace_403
    assert data.foo_bar_baz == "something"
    assert data.model_dump() == camel
    assert data.model_dump(by_alias=False) == snake
    assert json.loads(data.model_dump_json()) == camel
    assert json.loads(data.model_dump_json(by_alias=False)) == snake

    snake_data = TestModel.model_validate(snake)
    assert data.minimum_lifetime == 10
    assert not data.replace_403
    assert data.foo_bar_baz == "something"
    assert snake_data.model_dump() == data.model_dump()
    assert snake_data.model_dump_json() == data.model_dump_json()


def test_validate_exactly_one_of() -> None:
    class Model(BaseModel):
        foo: int | None = None
        bar: int | None = None
        baz: int | None = None

        _validate_type = model_validator(mode="after")(
            validate_exactly_one_of("foo", "bar", "baz")
        )

    Model.model_validate({"foo": 4, "bar": None})
    Model.model_validate({"baz": 4})
    Model.model_validate({"bar": 4})
    Model.model_validate({"foo": None, "bar": 4})

    with pytest.raises(ValidationError) as excinfo:
        Model.model_validate({"foo": 4, "bar": 3, "baz": None})
    assert "only one of foo, bar, and baz may be given" in str(excinfo.value)

    with pytest.raises(ValidationError) as excinfo:
        Model.model_validate({"foo": None, "baz": None})
    assert "one of foo, bar, and baz must be given" in str(excinfo.value)

    class TwoModel(BaseModel):
        foo: int | None = None
        bar: int | None = None

        _validate_type = model_validator(mode="after")(
            validate_exactly_one_of("foo", "bar")
        )

    with pytest.raises(ValidationError) as excinfo:
        TwoModel.model_validate({"foo": 3, "bar": 4})
    assert "only one of foo and bar may be given" in str(excinfo.value)

    with pytest.raises(ValidationError) as excinfo:
        TwoModel.model_validate({})
    assert "one of foo and bar must be given" in str(excinfo.value)



"""Tests for VOSI functionality provided by the UWS library."""

from __future__ import annotations

import pytest
from httpx import AsyncClient
from vo_models.vosi.availability import Availability

AVAILABILITY = """
<vosi:availability xmlns:vosi="http://www.ivoa.net/xml/VOSIAvailability/v1.0">
  <vosi:available>true</vosi:available>
</vosi:availability>
"""


@pytest.mark.asyncio
async def test_availability(client: AsyncClient) -> None:
    r = await client.get("/test/availability")
    assert r.status_code == 200
    assert Availability.from_xml(r.text) == Availability.from_xml(AVAILABILITY)



from __future__ import annotations

import gidgethub.apps
import httpx
from gidgethub.httpx import GitHubAPI


class GitHubAppClientFactory:
    """Factory for creating GitHub App clients authenticated either as an app
    or as an installation of that app.

    Parameters
    ----------
    id
        The GitHub App ID.
    key
        The GitHub App private key.
    name
        The GitHub App name. This identifies the app in the user agent string,
        and is typically the name of the GitHub repository the app is built
        from (e.g. ``lsst-sqre/times-square``).
    http_client
        The httpx client.

    Notes
    -----
    Gidgethub treats the application ID and installation ID as strings, but
    GitHub's API appears to return them as integers. This class expects them
    to be integers and converts them to strings when calling Gidgethub.
    """

    def __init__(
        self, *, id: int, key: str, name: str, http_client: httpx.AsyncClient
    ) -> None:
        self.app_id = id
        self.app_key = key
        self.app_name = name
        self._http_client = http_client

    def get_app_jwt(self) -> str:
        """Create the GitHub App's JWT based on application configuration.

        This token is for authenticating as the GitHub App itself, as opposed
        to an installation of the app.

        Returns
        -------
        str
            The JWT token.
        """
        return gidgethub.apps.get_jwt(
            app_id=str(self.app_id), private_key=self.app_key
        )

    def _create_client(self, *, oauth_token: str | None = None) -> GitHubAPI:
        return GitHubAPI(
            self._http_client, self.app_name, oauth_token=oauth_token
        )

    def create_anonymous_client(self) -> GitHubAPI:
        """Create an anonymous client.

        Returns
        -------
        gidgethub.httpx.GitHubAPI
            The anonymous client.
        """
        return self._create_client()

    async def create_installation_client(
        self, installation_id: int
    ) -> GitHubAPI:
        """Create a client authenticated as an installation of the GitHub App
        for a specific repository or organization.

        Parameters
        ----------
        installation_id
            The installation ID. This can be retrieved from the
            ``installation.id`` field of a webhook payload or from the
            ``id`` field of the ``GET "/repos/{owner}/{repo}/installation"``
            GitHub endpoint.

        Returns
        -------
        gidgethub.httpx.GitHubAPI
            The installation client.
        """
        anon_client = self.create_anonymous_client()
        token_info = await gidgethub.apps.get_installation_access_token(
            anon_client,
            installation_id=str(installation_id),
            app_id=str(self.app_id),
            private_key=self.app_key,
        )
        return self._create_client(oauth_token=token_info["token"])

    async def create_installation_client_for_repo(
        self, owner: str, repo: str
    ) -> GitHubAPI:
        """Create a client authenticated as an installation of the GitHub App
        for a specific repository or organization.

        Parameters
        ----------
        owner
            The owner of the repository.
        repo
            The repository name.

        Returns
        -------
        gidgethub.httpx.GitHubAPI
            The installation client.
        """
        app_jwt = self.get_app_jwt()
        anon_client = self.create_anonymous_client()
        installation_data = await anon_client.getitem(
            "/repos/{owner}/{repo}/installation",
            url_vars={"owner": owner, "repo": repo},
            jwt=app_jwt,
        )
        installation_id = installation_data["id"]
        return await self.create_installation_client(installation_id)



"""Gafaelfawr authentication dependencies."""

from typing import Annotated

from fastapi import Depends, Header
from structlog.stdlib import BoundLogger

from .logger import logger_dependency

__all__ = [
    "auth_delegated_token_dependency",
    "auth_dependency",
    "auth_logger_dependency",
]


async def auth_dependency(
    x_auth_request_user: Annotated[str, Header(include_in_schema=False)],
) -> str:
    """Retrieve authentication information from HTTP headers.

    Intended for use with applications protected by Gafaelfawr, this retrieves
    authentication information from headers added to the incoming request by
    the Gafaelfawr ``auth_request`` NGINX subhandler.
    """
    return x_auth_request_user


async def auth_delegated_token_dependency(
    x_auth_request_token: Annotated[str, Header(include_in_schema=False)],
) -> str:
    """Retrieve Gafaelfawr delegated token from HTTP headers.

    Intended for use with applications protected by Gafaelfawr, this retrieves
    a delegated token from headers added to the incoming request by the
    Gafaelfawr ``auth_request`` NGINX subhandler. The delegated token can be
    used to make requests to other services on the user's behalf. See `the
    Gafaelfawr documentation
    <https://gafaelfawr.lsst.io/user-guide/gafaelfawringress.html#requesting-delegated-tokens>`__
    for more details.
    """
    return x_auth_request_token


async def auth_logger_dependency(
    user: Annotated[str, Depends(auth_dependency)],
    logger: Annotated[BoundLogger, Depends(logger_dependency)],
) -> BoundLogger:
    """Logger bound to the authenticated user."""
    return logger.bind(user=user)



"""Metrics exceptions."""

__all__ = [
    "DuplicateEventError",
    "EventManagerUninitializedError",
    "EventManagerUsageError",
    "KafkaTopicError",
    "UnsupportedAvroSchemaError",
]


class EventManagerUsageError(Exception):
    """These exceptions should be raised even in abandonable methods.

    They represent application errors that can and should be fixed, vs. errors
    with infrastructure outside of the application's control, like the
    underlying Kafka infrastructure.
    """


class UnsupportedAvroSchemaError(EventManagerUsageError):
    """Event model is not serializable to Avro."""


class EventManagerUninitializedError(EventManagerUsageError):
    """An attempt to create a publisher after manager has been initialized."""


class DuplicateEventError(EventManagerUsageError):
    """Two publishers were registered with the same name."""

    def __init__(self, name: str) -> None:
        super().__init__(
            f"{name}: you have already created an event with this "
            " name. Events must have unique names within this "
            " application."
        )


class KafkaTopicError(EventManagerUsageError):
    """A topic does not exist in Kafka, or we don't have access to it."""

    def __init__(self, topic: str) -> None:
        super().__init__(
            f"Topic: {topic} does not exist in Kafka, or we don't have the"
            " permissions to access it. See https://safir.lsst.io/metrics"
            " for how to provision Kafka topics, access, and other "
            " infrastructure for publishing metrics events."
        )



"""Alembic migration environment.

Keep in sync with the instructions in the Safir user guide.
"""

from alembic import context

from safir.database import run_migrations_offline, run_migrations_online
from safir.logging import configure_alembic_logging, configure_logging
from tests.support.alembic import BaseV1, config

# Configure structlog.
configure_logging(name="tests", log_level=config.log_level)
configure_alembic_logging()

# Run the migrations.
if context.is_offline_mode():
    run_migrations_offline(BaseV1.metadata, config.database_url)
else:
    run_migrations_online(
        BaseV1.metadata,
        config.database_url,
        config.database_password,
    )



"""Helpers for Sentry support in SlackExceptions."""

from dataclasses import dataclass, field
from typing import Any

__all__ = ["SentryEventInfo"]


@dataclass
class SentryEventInfo:
    """A collection of metadata to add to Sentry events.

    The overrideable `~SlackException.to_sentry` method returns an object of
    this type.
    """

    tags: dict[str, str] = field(default_factory=dict)
    """Short key-value pairs to add to a Sentry event.

    Use tags for small values that you would like to search by and aggregate
    over when analyzing multiple Sentry events in the Sentry UI.
    """

    contexts: dict[str, dict[str, Any]] = field(default_factory=dict)
    """Detailed information to add to a Sentry event.

    You can not search by context values, but you can store more data in them.
    """

    attachments: dict[str, Any] = field(default_factory=dict)
    """Long strings of text, or other files, to add to a Sentry event.

    Attachments can hold the most text, but are the hardest to view in the
    Sentry UI. They can also hold non-text files.
    """
    username: str | None = None



"""Pydantic models for GitHub v3 REST API resources."""

from __future__ import annotations

from base64 import b64decode
from enum import StrEnum

from pydantic import BaseModel, Field, HttpUrl

__all__ = [
    "GitHubBlobModel",
    "GitHubBranchCommitModel",
    "GitHubBranchModel",
    "GitHubCheckRunAnnotationLevel",
    "GitHubCheckRunConclusion",
    "GitHubCheckRunModel",
    "GitHubCheckRunOutput",
    "GitHubCheckRunPrInfoModel",
    "GitHubCheckRunStatus",
    "GitHubCheckSuiteConclusion",
    "GitHubCheckSuiteId",
    "GitHubCheckSuiteModel",
    "GitHubCheckSuiteStatus",
    "GitHubPullRequestHeadModel",
    "GitHubPullRequestModel",
    "GitHubPullState",
    "GitHubRepoOwnerModel",
    "GitHubRepositoryModel",
    "GitHubUserModel",
]


class GitHubRepoOwnerModel(BaseModel):
    """A Pydantic model for the ``owner`` field found in repository objects.

    https://docs.github.com/en/rest/repos/repos#get-a-repository
    """

    login: str = Field(
        title="Login name",
        description=(
            "Login name of the owner (either a user or an organization)."
        ),
        examples=["lsst-sqre"],
    )


class GitHubUserModel(BaseModel):
    """A Pydantic model for the ``user`` field found in GitHub API resources.

    This contains brief (public) info about a user.
    """

    login: str = Field(title="Login name", description="GitHub username.")

    html_url: HttpUrl = Field(
        title="Profile URL", description="Homepage for the user on GitHub."
    )

    url: HttpUrl = Field(
        title="API URL",
        description="URL for the user's resource in the GitHub API.",
    )

    avatar_url: HttpUrl = Field(
        title="Avatar image URL", description="URL to the user's avatar."
    )


class GitHubRepositoryModel(BaseModel):
    """A Pydantic model for the ``repository`` field, often found in webhook
    payloads.

    https://docs.github.com/en/rest/repos/repos#get-a-repository
    """

    name: str = Field(
        title="Repository name",
        description="Excludes owner prefix.",
        examples=["times-square-demo"],
    )

    full_name: str = Field(
        title="Full name",
        description=(
            "Full name, including owner prefix "
            "(e.g. ``lsst-sqre/times-square-demo``).)"
        ),
        examples=["lsst-sqre/times-square-demo"],
    )

    owner: GitHubRepoOwnerModel = Field(description="The repository's owner.")

    default_branch: str = Field(
        description="The default branch (e.g. main).", examples=["main"]
    )

    html_url: HttpUrl = Field(
        description="URL of the repository for browsers.",
        examples=["https://github.com/lsst-sqre/times-square-demo"],
    )

    branches_url: str = Field(
        description="URI template for the repo's branches endpoint.",
        examples=[
            "https://github.com/lsst-sqre/times-square-demo/branches{/branch}"
        ],
    )

    contents_url: str = Field(
        description="URI template for the contents endpoint.",
        examples=[
            "https://github.com/lsst-sqre/times-square-demo/contents/{+path}"
        ],
    )

    trees_url: str = Field(
        description="URI template for the Git tree API.",
        examples=[
            "https://github.com/lsst-sqre/times-square-demo/git/trees{/sha}"
        ],
    )

    blobs_url: str = Field(
        description="URI template for the Git blobs API.",
        examples=[
            "https://github.com/lsst-sqre/times-square-demo/git/blobs{/sha}"
        ],
    )


class GitHubPullState(StrEnum):
    """The state of a GitHub pull request (PR).

    https://docs.github.com/en/rest/pulls/pulls#get-a-pull-request
    """

    open = "open"
    """The PR is open."""

    closed = "closed"
    """The PR is closed."""


class GitHubPullRequestHeadModel(BaseModel):
    """A Pydantic model for the last commit to the head branch of a PR."""

    sha: str = Field(description="Last commit SHA of the PR head branch.")


class GitHubPullRequestModel(BaseModel):
    """A Pydantic model for a GitHub Pull Request.

    This is also the ``pull_request`` field inside the
    `~safir.github.webhooks.GitHubPullRequestEventModel`.

    https://docs.github.com/en/rest/pulls/pulls#get-a-pull-request
    """

    html_url: HttpUrl = Field(description="Web URL of the PR.")

    number: int = Field(description="Pull request number.")

    title: str = Field(description="Title of the PR.")

    state: GitHubPullState = Field(
        description="Whether the PR is opened or closed."
    )

    draft: bool = Field(description="True if the PR is a draft.")

    merged: bool = Field(description="True if the PR is merged.")

    user: GitHubUserModel = Field(description="The user that opened the PR.")

    head: GitHubPullRequestHeadModel = Field(
        description="The last commit to the head branch of the PR."
    )


class GitHubBranchCommitModel(BaseModel):
    """A Pydantic model for the commit field found in `GitHubBranchModel`."""

    sha: str = Field(description="Git commit SHA.")

    url: HttpUrl = Field(description="URL for commit resource.")


class GitHubBranchModel(BaseModel):
    """A Pydantic model for a GitHub branch.

    https://docs.github.com/en/rest/branches/branches#get-a-branch
    """

    name: str = Field(description="Branch name (e.g. main)", examples=["main"])

    commit: GitHubBranchCommitModel = Field(description="HEAD commit info.")


class GitHubBlobModel(BaseModel):
    """A Pydantic model for a blob, returned by the GitHub blob endpoint.

    See https://docs.github.com/en/rest/git/blobs#get-a-blob
    """

    content: str = Field(
        description=(
            "The blob's encoded content. Use the `decode` method to decode."
        )
    )

    encoding: str = Field(description="Content encoding (typically base64).")

    url: HttpUrl = Field(description="API URL of this resource.")

    sha: str = Field(description="Git SHA of tree object.")

    size: int = Field(description="Size of the content in bytes.")

    def decode(self) -> str:
        """Decode the `content` field.

        Currently supports these encodings:

        - base64

        Returns
        -------
        str
            The decoded content.
        """
        if self.encoding == "base64":
            return b64decode(self.content).decode()
        else:
            raise NotImplementedError(
                f"GitHub blob content encoding {self.encoding} "
                f"is unknown by GitHubBlobModel for url {self.url}"
            )


class GitHubCheckSuiteStatus(StrEnum):
    """The status of a GitHub check suite."""

    queued = "queued"
    """The check suite is queued."""

    in_progress = "in_progress"
    """The check suite is in progress."""

    completed = "completed"
    """The check suite has completed."""


class GitHubCheckSuiteConclusion(StrEnum):
    """The conclusion state of a GitHub check suite."""

    success = "success"
    """The check suite has succeeded."""

    failure = "failure"
    """The check suite has failed."""

    neutral = "neutral"
    """The check suite has a neutral outcome, perhaps because the check was
    skipped.
    """

    cancelled = "cancelled"
    """The check suite was cancelled."""

    timed_out = "timed_out"
    """The check suite timed out."""

    action_required = "action_required"
    """The check suite requires an action to be taken before it can
    continue.
    """

    stale = "stale"
    """The check suite is stale."""


class GitHubCheckSuiteModel(BaseModel):
    """A Pydantic model for the ``check_suite`` field in a ``check_suite``
    webhook (`~safir.github.webhooks.GitHubCheckSuiteEventModel`).
    """

    id: int = Field(description="Identifier for this check run.")

    head_branch: str = Field(
        description="Name of the branch the changes are on.",
    )

    head_sha: str = Field(
        description="The SHA of the most recent commit for this check suite.",
    )

    url: HttpUrl = Field(
        description="GitHub API URL for the check suite resource."
    )

    status: GitHubCheckSuiteStatus = Field(
        description="The status of the check suite."
    )

    conclusion: GitHubCheckSuiteConclusion | None = Field(
        description="The conclusion of the check suite."
    )

    pull_requests: list[GitHubCheckRunPrInfoModel] = Field(
        description="A list of pull requests associated with this check suite."
    )


class GitHubCheckRunStatus(StrEnum):
    """The check run status."""

    queued = "queued"
    """The check run is queued."""

    in_progress = "in_progress"
    """The check run is in progress."""

    completed = "completed"
    """The check run has completed."""


class GitHubCheckRunConclusion(StrEnum):
    """The check run conclusion state."""

    success = "success"
    """The check run has succeeded."""

    failure = "failure"
    """The check run has failed."""

    neutral = "neutral"
    """The check run has a neutral outcome, perhaps because the check was
    skipped.
    """

    cancelled = "cancelled"
    """The check run was cancelled."""

    timed_out = "timed_out"
    """The check run timed out."""

    action_required = "action_required"
    """The check run requires an action to be taken before it can continue."""

    stale = "stale"
    """The check run is stale."""


class GitHubCheckRunAnnotationLevel(StrEnum):
    """The level of a check run output annotation."""

    notice = "notice"
    """A notice annotation."""

    warning = "warning"
    """A warning annotation."""

    failure = "failure"
    """An annotation that indicates a failure."""


class GitHubCheckSuiteId(BaseModel):
    """Brief information about a check suite in the `GitHubCheckRunModel`."""

    id: int = Field(description="Check suite ID")


class GitHubCheckRunOutput(BaseModel):
    """Check run output report."""

    title: str | None = Field(None, description="Title of the report")

    summary: str | None = Field(
        None, description="Summary information (markdown formatted)."
    )

    text: str | None = Field(None, description="Extended report (markdown)")


class GitHubCheckRunPrInfoModel(BaseModel):
    """A Pydantic model of the ``pull_requests[]`` items in a check run
    GitHub API model (`GitHubCheckRunModel`).

    https://docs.github.com/en/rest/checks/runs#get-a-check-run
    """

    url: HttpUrl = Field(description="GitHub API URL for this pull request.")

    number: int = Field(
        description="The number that identifies the pull request."
    )


class GitHubCheckRunModel(BaseModel):
    """A Pydantic model for the "check_run" field in a check_run webhook
    payload (`~safir.github.webhooks.GitHubCheckRunEventModel`).
    """

    id: int = Field(description="Identifier for this check run.")

    external_id: str | None = Field(
        description="Identifier set by the check runner."
    )

    head_sha: str = Field(
        title="Head sha",
        description="The SHA of the most recent commit for this check suite.",
    )

    status: GitHubCheckRunStatus = Field(
        description="Status of the check run."
    )

    conclusion: GitHubCheckRunConclusion | None = Field(
        None, description="Conclusion status, if completed."
    )

    name: str = Field(description="Name of the check run.")

    url: HttpUrl = Field(description="URL of the check run API resource.")

    html_url: HttpUrl = Field(description="URL of the check run webpage.")

    check_suite: GitHubCheckSuiteId = Field(
        description="Brief information about the check suite."
    )

    output: GitHubCheckRunOutput | None = Field(
        None, title="Output", description="Check run output, if available."
    )

    pull_requests: list[GitHubCheckRunPrInfoModel] = Field(
        default_factory=list,
        description="List of pull requests associated with this check run.",
    )



"""Test Slack client."""

from __future__ import annotations

import pickle
from textwrap import dedent
from unittest.mock import ANY

import pytest
import respx
import sentry_sdk
import structlog
from httpx import AsyncClient, HTTPError, Response
from pydantic import ValidationError

from safir.slack.blockkit import (
    SlackBaseField,
    SlackCodeBlock,
    SlackCodeField,
    SlackException,
    SlackMessage,
    SlackTextBlock,
    SlackTextField,
    SlackWebException,
)
from safir.slack.webhook import SlackWebhookClient
from safir.testing.sentry import Captured
from safir.testing.slack import MockSlackWebhook


def test_message() -> None:
    message = SlackMessage(
        message="This is some *Slack message*  \n  ",
        fields=[
            SlackTextField(heading="Some text", text="Value of the field   "),
            SlackCodeField(heading="Some code", code="Here is\nthe code\n"),
        ],
        blocks=[SlackTextBlock(heading="Log", text="Some\nlong\nlog")],
        attachments=[
            SlackCodeBlock(heading="Backtrace", code="Some\nbacktrace"),
            SlackTextBlock(heading="Essay", text="Blah blah blah"),
        ],
    )
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "This is some *Slack message*",
                    "verbatim": True,
                },
            },
            {
                "type": "section",
                "fields": [
                    {
                        "type": "mrkdwn",
                        "text": "*Some text*\nValue of the field",
                        "verbatim": True,
                    },
                    {
                        "type": "mrkdwn",
                        "text": "*Some code*\n```\nHere is\nthe code\n```",
                        "verbatim": True,
                    },
                ],
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "*Log*\nSome\nlong\nlog",
                    "verbatim": True,
                },
            },
        ],
        "attachments": [
            {
                "blocks": [
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": "*Backtrace*\n```\nSome\nbacktrace\n```",
                            "verbatim": True,
                        },
                    },
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": "*Essay*\nBlah blah blah",
                            "verbatim": True,
                        },
                    },
                ]
            }
        ],
    }

    message = SlackMessage(message="Single line message", verbatim=False)
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Single line message",
                    "verbatim": False,
                },
            }
        ]
    }

    message = SlackMessage(
        message="Message with <special> & one `field`",
        fields=[SlackTextField(heading="Something", text="Blah")],
    )
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Message with &lt;special&gt; &amp; one `field`",
                    "verbatim": True,
                },
            },
            {
                "type": "section",
                "fields": [
                    {
                        "type": "mrkdwn",
                        "text": "*Something*\nBlah",
                        "verbatim": True,
                    },
                ],
            },
            {"type": "divider"},
        ]
    }

    message = SlackMessage(
        message="Message with one block",
        blocks=[SlackTextBlock(heading="Something", text="Blah")],
    )
    assert message.to_slack() == {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Message with one block",
                    "verbatim": True,
                },
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "*Something*\nBlah",
                    "verbatim": True,
                },
            },
            {"type": "divider"},
        ]
    }


def test_validation() -> None:
    """Test errors caught by validation."""
    fields: list[SlackBaseField] = [
        SlackTextField(heading="Something", text="foo")
    ] * 11
    message = SlackMessage(message="Ten fields", fields=fields[:10])
    assert len(message.fields) == 10
    with pytest.raises(ValidationError):
        SlackMessage(message="Eleven fields", fields=fields)


def test_block_truncation() -> None:
    """Test truncating attachments at Slack limits."""
    block = SlackTextBlock(heading="Something", text="a" * 3000)
    length = 3000 - len("*Something*\n\n... truncated ...")
    assert block.to_slack()["text"] == (
        "*Something*\n" + "a" * length + "\n... truncated ..."
    )

    block = SlackTextBlock(heading="Something", text="abcde\n" * 500)
    length = int((3001 - len("*Something*\n\n... truncated ...")) / 6)
    assert block.to_slack()["text"] == (
        "*Something*\n" + "abcde\n" * length + "... truncated ..."
    )

    cblock = SlackCodeBlock(heading="Else", code="a" * 3000)
    length = 3000 - len("*Else*\n```\n... truncated ...\n\n```")
    assert cblock.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n" + "a" * length + "\n```"
    )

    cblock = SlackCodeBlock(heading="Else", code="abcde\n" * 500)
    length = int((3001 - len("*Else*\n```\n... truncated ...\n\n```")) / 6)
    assert cblock.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n"
        + ("abcde\n" * length).strip()
        + "\n```"
    )


def test_field_truncation() -> None:
    """Test truncating fields at Slack limits."""
    field = SlackTextField(heading="Something", text="a" * 2000)
    length = 2000 - len("*Something*\n\n... truncated ...")
    assert field.to_slack()["text"] == (
        "*Something*\n" + "a" * length + "\n... truncated ..."
    )

    field = SlackTextField(heading="Something", text="abcdefg\n" * 250)
    length = int((2001 - len("*Something*\n\n... truncated ...")) / 8)
    assert field.to_slack()["text"] == (
        "*Something*\n" + "abcdefg\n" * length + "... truncated ..."
    )

    cfield = SlackCodeField(heading="Else", code="a" * 2000)
    length = 2000 - len("*Else*\n```\n... truncated ...\n\n```")
    assert cfield.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n" + "a" * length + "\n```"
    )

    cfield = SlackCodeField(heading="Else", code="abcdefg\n" * 250)
    length = int((2001 - len("*Else*\n```\n... truncated ...\n\n```")) / 8)
    assert cfield.to_slack()["text"] == (
        "*Else*\n```\n... truncated ...\n"
        + ("abcdefg\n" * length).strip()
        + "\n```"
    )


def test_message_truncation() -> None:
    """Text truncating the main part of a Slack message."""
    message = SlackMessage(message="a" * 3000)
    assert message.to_slack()["blocks"][0]["text"]["text"] == "a" * 3000
    message = SlackMessage(message="a" * 3001)
    length = 3000 - len("\n... truncated ...")
    assert message.to_slack()["blocks"][0]["text"]["text"] == (
        "a" * length + "\n... truncated ..."
    )


@pytest.mark.asyncio
async def test_exception(mock_slack: MockSlackWebhook) -> None:
    logger = structlog.get_logger(__file__)
    slack = SlackWebhookClient(mock_slack.url, "App", logger)

    class SomeError(SlackException):
        pass

    try:
        raise SomeError("Some error", "someuser")
    except SlackException as e:
        await slack.post_exception(e)

    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: Some error",
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nSomeError",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": ANY,
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": "*User*\nsomeuser",
                            "verbatim": True,
                        },
                    ],
                },
                {"type": "divider"},
            ],
        }
    ]


class SlackSubclassException(SlackException):
    """Subclsas for testing pickling."""

    def __init__(self) -> None:
        super().__init__("Some error", "username")


def test_exception_pickling() -> None:
    """Test that a `SlackException` can be pickled and unpickled.

    Errors that may be raised by backend workers in an arq queue must support
    pickling so that they can be passed correctly to other workers or the
    frontend.
    """
    exc = SlackException("some message", "username")
    pickled_exc = pickle.loads(pickle.dumps(exc))
    assert exc.to_slack() == pickled_exc.to_slack()

    # Try the same with a derived class with a different number of constructor
    # arguments.
    subexc = SlackSubclassException()
    pickled_subexc = pickle.loads(pickle.dumps(subexc))
    assert subexc.to_slack() == pickled_subexc.to_slack()


@pytest.mark.asyncio
async def test_web_exception(
    respx_mock: respx.Router, mock_slack: MockSlackWebhook
) -> None:
    logger = structlog.get_logger(__file__)
    slack = SlackWebhookClient(mock_slack.url, "App", logger)

    class SomeError(SlackWebException):
        pass

    respx_mock.get("https://example.org/").mock(return_value=Response(404))
    exception = None
    try:
        async with AsyncClient() as client:
            r = await client.get("https://example.org/")
            r.raise_for_status()
    except HTTPError as e:
        exception = SomeError.from_exception(e)
        assert str(exception) == "Status 404 from GET https://example.org/"
        await slack.post_exception(exception)

    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: " + str(exception),
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nSomeError",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": ANY,
                            "verbatim": True,
                        },
                    ],
                },
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "*URL*\nGET https://example.org/",
                        "verbatim": True,
                    },
                },
                {"type": "divider"},
            ],
        }
    ]


@pytest.mark.asyncio
async def test_web_exception_sentry(
    respx_mock: respx.Router,
    sentry_combo_items: Captured,
) -> None:
    class SomeError(SlackWebException):
        pass

    respx_mock.get("https://example.org/").mock(
        return_value=Response(404, text="some response body")
    )
    exception = None
    try:
        async with AsyncClient() as client:
            r = await client.get("https://example.org/")
            r.raise_for_status()
    except HTTPError as e:
        exception = SomeError.from_exception(e)
        stringified = dedent("""\
            Status 404 from GET https://example.org/
            Body:
            some response body
        """)
        assert str(exception) == stringified
        sentry_sdk.capture_exception(exception)

    (error,) = sentry_combo_items.errors
    assert error["tags"]["httpx_request_method"] == "GET"
    assert error["tags"]["httpx_request_url"] == "https://example.org/"

    (attachment,) = sentry_combo_items.attachments
    assert attachment.filename == "httpx_response_body"
    assert attachment.bytes.decode() == "some response body"



"""Representation of a UWS application."""

from __future__ import annotations

from typing import Any

from fastapi import APIRouter, FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import PlainTextResponse
from structlog.stdlib import BoundLogger

try:
    from safir.arq import ArqQueue, WorkerSettings
    from safir.arq.uws import UWS_QUEUE_NAME
except ImportError as e:
    raise ImportError(
        "The safir.uws module requires the uws extra. "
        "Install it with `pip install safir[uws]`."
    ) from e
from safir.middleware.ivoa import (
    CaseInsensitiveFormMiddleware,
    CaseInsensitiveQueryMiddleware,
)

from ._config import UWSConfig
from ._constants import UWS_DATABASE_TIMEOUT
from ._dependencies import uws_dependency
from ._exceptions import UWSError
from ._handlers import (
    install_async_post_handler,
    install_availability_handler,
    install_sync_get_handler,
    install_sync_post_handler,
    uws_router,
)
from ._workers import (
    close_uws_worker_context,
    create_uws_worker_context,
    uws_job_completed,
    uws_job_started,
)

__all__ = ["UWSApplication"]


async def _uws_error_handler(
    request: Request, exc: UWSError
) -> PlainTextResponse:
    response = f"{exc.error_code}: {exc!s}\n"
    if exc.detail:
        response += "\n{exc.detail}"
    return PlainTextResponse(response, status_code=exc.status_code)


async def _usage_handler(
    request: Request, exc: RequestValidationError
) -> PlainTextResponse:
    return PlainTextResponse(f"UsageError\n\n{exc!s}", status_code=422)


class UWSApplication:
    """Glue between a FastAPI application and the UWS implementation.

    An instance of this class should be created during construction of the
    service that will use the UWS layer. It provides methods to build route
    handlers, install error handlers, and build the UWS database
    worker. Construction of the backend worker that does the work of the
    service is handled separately so that it can have minimal dependencies.

    Parameters
    ----------
    config
        UWS configuration.
    """

    def __init__(self, config: UWSConfig) -> None:
        self._config = config

    def build_worker(self, logger: BoundLogger) -> WorkerSettings:
        """Construct an arq worker configuration for the UWS worker.

        All UWS job status and results must be stored in the underlying
        database (via Wobbly), since the API serves job information from
        there. To minimize dependencies for the worker, which may (for
        example) pin its own version of SQLAlchemy that may not be compatible
        with that used by the application, the actual worker is not
        responsible for storing the results in the database. Instead, it
        returns results via arq, which temporarily puts them in Redis then
        uses ``on_job_start`` and ``after_job_end`` to notify a different
        queue. Those results are recovered and stored in the database (via
        Wobbly) by a separate arq worker.

        This function defines that database worker. It returns a class
        suitable for assigning to a module variable and referencing as the
        argument to the :command:`arq` command-line tool to start the worker.

        Parameters
        ----------
        logger
            Logger to use for messages.
        """

        async def startup(ctx: dict[Any, Any]) -> None:
            ctx.update(await create_uws_worker_context(self._config, logger))

        async def shutdown(ctx: dict[Any, Any]) -> None:
            await close_uws_worker_context(ctx)

        # Running 10 jobs simultaneously is the arq default as of arq 0.26.0
        # and seems reasonable for database workers.
        return WorkerSettings(
            functions=[uws_job_started, uws_job_completed],
            redis_settings=self._config.arq_redis_settings,
            job_completion_wait=UWS_DATABASE_TIMEOUT,
            job_timeout=UWS_DATABASE_TIMEOUT,
            max_jobs=10,
            queue_name=UWS_QUEUE_NAME,
            on_startup=startup,
            on_shutdown=shutdown,
        )

    async def initialize_fastapi(self) -> None:
        """Initialize the UWS subsystem for FastAPI applications.

        This must be called before any UWS routes are accessed, normally from
        the lifespan function of the FastAPI application.
        """
        await uws_dependency.initialize(self._config)

    def install_error_handlers(self, app: FastAPI) -> None:
        """Install error handlers that follow DALI and UWS conventions.

        This method must be called during application setup for any FastAPI
        app using the UWS layer for correct error message handling. This will
        change the error response for all parameter validation errors from
        FastAPI.

        Currently these error handlers return ``text/plain`` errors. VOTable
        errors may be a better choice, but revision 1.0 of the SODA standard
        only allows ``text/plain`` errors for sync routes.
        """
        app.exception_handler(UWSError)(_uws_error_handler)
        app.exception_handler(RequestValidationError)(_usage_handler)

    def install_handlers(self, router: APIRouter) -> None:
        """Install the route handlers for the service.

        This method will always install a POST handler at the root of the
        router that creates an async job, and handlers under ``/jobs`` that
        implement the UWS protocol for managing those jobs. If
        ``sync_post_route`` is set in the `~safir.uws.UWSConfig` that this
        application was configured with, a POST handler for ``/sync`` to
        create a sync job will be added. If ``sync_get_route`` is set, a GET
        handler for ``/sync`` to create a sync job will be added.
        """
        router.include_router(uws_router, prefix="/jobs")
        install_availability_handler(router)
        if route := self._config.sync_get_route:
            install_sync_get_handler(router, route)
        if route := self._config.sync_post_route:
            install_sync_post_handler(router, route)

        # This handler must be installed directly on the provided router. Do
        # not install it on the UWS router before include_router. The process
        # of copying handlers done by include_router loses the dependency
        # information from the async post handler and causes it to not see any
        # of its job parameters.
        #
        # This is probably because the dependency is a dynamic function not
        # known statically, which may confuse the handler copying code in
        # FastAPI. This problem was last verified in FastAPI 0.111.0.
        install_async_post_handler(router, self._config.async_post_route)

    def install_middleware(self, app: FastAPI) -> None:
        """Install FastAPI middleware needed by UWS.

        UWS unfortunately requires that the key portion of query and form
        parameters be case-insensitive, so UWS FastAPI applications need to
        add custom middleware to lowercase parameter keys. This method does
        that.

        Parameters
        ----------
        app
            FastAPI app.
        """
        app.add_middleware(CaseInsensitiveFormMiddleware)
        app.add_middleware(CaseInsensitiveQueryMiddleware)

    def override_arq_queue(self, arq_queue: ArqQueue) -> None:
        """Change the arq used by the FastAPI route handlers.

        This method is probably only useful for the test suite.

        Parameters
        ----------
        arq_queue
            New arq queue.
        """
        uws_dependency.override_arq_queue(arq_queue)

    async def shutdown_fastapi(self) -> None:
        """Shut down the UWS subsystem for FastAPI applications.

        This should be called during application shutdown, normally from the
        lifespan function of the FastAPI application. Currently, this does
        nothing, but it remains as a hook in case some shutdown is required in
        the future.
        """



"""Tests for the UWS job manipulation handlers.

These tests don't assume any given application, and therefore don't use the
API to create a job, instead inserting it directly via the UWSService.
"""

from __future__ import annotations

import asyncio
from datetime import UTC, datetime, timedelta
from unittest.mock import ANY

import pytest
from arq.constants import default_queue_name
from arq.jobs import JobStatus
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from vo_models.uws import JobSummary, Results

from safir.arq import MockArqQueue
from safir.arq.uws import WorkerJobInfo, WorkerResult
from safir.datetime import current_datetime, isodatetime
from safir.testing.uws import MockUWSJobRunner, assert_job_summary_equal
from safir.uws import Job, UWSConfig
from safir.uws._dependencies import UWSFactory

from ..support.uws import SimpleParameters, SimpleXmlParameters

PENDING_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>{}</uws:jobId>
  <uws:runId>some-run-id</uws:runId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>{}</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:executionDuration>{}</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Jane</uws:parameter>
  </uws:parameters>
</uws:job>
"""

FINISHED_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:runId>some-run-id</uws:runId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>COMPLETED</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:endTime>{}</uws:endTime>
  <uws:executionDuration>{}</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Jane</uws:parameter>
  </uws:parameters>
  <uws:results>
    <uws:result id="cutout" xlink:href="https://example.com/some/path"\
 mime-type="application/fits"/>
  </uws:results>
</uws:job>
"""

ABORTED_PENDING_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>{}</uws:jobId>
  <uws:runId>some-run-id</uws:runId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>ABORTED</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Jane</uws:parameter>
  </uws:parameters>
</uws:job>
"""

ABORTED_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>{}</uws:jobId>
  <uws:runId>some-run-id</uws:runId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>ABORTED</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:endTime>{}</uws:endTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Jane</uws:parameter>
  </uws:parameters>
</uws:job>
"""

JOB_PARAMETERS = """
<uws:parameters
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:parameter id="name">Jane</uws:parameter>
</uws:parameters>
"""

JOB_RESULTS = """
<uws:results
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:result id="cutout" xlink:href="https://example.com/some/path"\
 mime-type="application/fits"/>
</uws:results>
"""


@pytest.mark.asyncio
async def test_job_run(
    client: AsyncClient,
    test_token: str,
    test_username: str,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    uws_config: UWSConfig,
) -> None:
    job_service = uws_factory.create_job_service()

    # Create the job.
    r = await client.post(
        "/test/jobs",
        data={"runid": "some-run-id", "name": "Jane"},
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"
    job = await job_service.get(test_token, "1")
    assert job.creation_time.microsecond == 0

    # Check the retrieval of the job configuration.
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "1",
            "PENDING",
            isodatetime(job.creation_time),
            "600",
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )

    # Try to put the job in an invalid phase.
    r = await client.post("/test/jobs/1/phase", data={"PHASE": "EXECUTING"})
    assert r.status_code == 422
    assert r.text.startswith("UsageError")

    # Start the job.
    r = await client.post(
        "/test/jobs/1/phase", data={"PHASE": "RUN"}, follow_redirects=True
    )
    assert r.status_code == 200
    assert r.url == "https://example.com/test/jobs/1"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "1",
            "QUEUED",
            isodatetime(job.creation_time),
            "600",
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )
    await runner.mark_in_progress(test_token, "1")

    # Check that the correct data was passed to the backend worker.
    metadata = await runner.get_job_metadata(test_token, "1")
    assert metadata.name == uws_config.worker
    assert metadata.args[0] == {"name": "Jane"}
    assert metadata.args[1] == WorkerJobInfo(
        job_id="1",
        user=test_username,
        token=test_token,
        timeout=ANY,
        run_id="some-run-id",
    )
    assert not metadata.kwargs
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=2) <= metadata.enqueue_time <= now
    assert metadata.status == JobStatus.in_progress
    assert metadata.queue_name == default_queue_name

    # Tell the queue the job is finished.
    results = [
        WorkerResult(
            result_id="cutout",
            url="s3://some-bucket/some/path",
            mime_type="application/fits",
        )
    ]
    job = await runner.mark_complete(test_token, "1", results)

    # Check the job results.
    assert job.start_time
    assert job.start_time.microsecond == 0
    assert job.end_time
    assert job.end_time.microsecond == 0
    assert job.end_time >= job.start_time >= job.creation_time
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        FINISHED_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            "600",
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )

    # Check that the phase is now correct.
    r = await client.get("/test/jobs/1/phase")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert r.text == "COMPLETED"

    # Retrieve them directly through the results resource.
    r = await client.get("/test/jobs/1/results")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    assert Results.from_xml(r.text) == Results.from_xml(JOB_RESULTS)

    # There should be no error message.
    r = await client.get("/test/jobs/1/error")
    assert r.status_code == 404


@pytest.mark.asyncio
async def test_job_abort(
    client: AsyncClient,
    test_token: str,
    runner: MockUWSJobRunner,
    arq_queue: MockArqQueue,
    uws_factory: UWSFactory,
    uws_config: UWSConfig,
) -> None:
    job_service = uws_factory.create_job_service()

    # Create the job.
    r = await client.post(
        "/test/jobs", data={"runid": "some-run-id", "name": "Jane"}
    )
    assert r.status_code == 303
    job = await job_service.get(test_token, "1")

    # Immediately abort the job.
    r = await client.post(
        "/test/jobs/1/phase", data={"PHASE": "ABORT"}, follow_redirects=True
    )
    assert r.status_code == 200
    assert r.url == "https://example.com/test/jobs/1"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        ABORTED_PENDING_JOB.format(
            "1",
            isodatetime(job.creation_time),
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )

    # Create a second job and start it running.
    r = await client.post(
        "/test/jobs",
        data={"runid": "some-run-id", "name": "Jane", "phase": "RUN"},
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/2"
    await runner.mark_in_progress(test_token, "2")

    # Abort that job.
    r = await client.post(
        "/test/jobs/2/phase", data={"PHASE": "ABORT"}, follow_redirects=True
    )
    assert r.status_code == 200
    assert r.url == "https://example.com/test/jobs/2"
    job = await job_service.get(test_token, "2")
    assert job.start_time
    assert job.end_time
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        ABORTED_JOB.format(
            "2",
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )
    job_result = await runner.get_job_result(test_token, "2")
    assert not job_result.success
    assert isinstance(job_result.result, asyncio.CancelledError)

    # Deleting a job should also abort it. Also test a weird capitalization of
    # the phase parameter and the POST form of the delete support.
    r = await client.post(
        "/test/jobs",
        data={"runid": "some-run-id", "name": "Jane", "PHAse": "RUN"},
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/3"
    await runner.mark_in_progress(test_token, "3")
    job = await job_service.get(test_token, "3")
    r = await client.post("/test/jobs/3", data={"action": "DELETE"})
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs"
    assert job.message_id
    job_result = await arq_queue.get_job_result(job.message_id)
    assert not job_result.success
    assert isinstance(job_result.result, asyncio.CancelledError)


@pytest.mark.asyncio
async def test_job_api(
    client: AsyncClient,
    test_token: str,
    test_username: str,
    uws_factory: UWSFactory,
) -> None:
    job_service = uws_factory.create_job_service()

    # Create the job.
    r = await client.post(
        "/test/jobs", data={"runid": "some-run-id", "name": "Jane"}
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"
    job = await job_service.get(test_token, "1")

    # Check the retrieval of the job configuration.
    destruction_time = job.creation_time + timedelta(seconds=24 * 60 * 60)
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "1",
            "PENDING",
            isodatetime(job.creation_time),
            "600",
            isodatetime(destruction_time),
        ),
    )

    # Check retrieving each part separately.
    r = await client.get("/test/jobs/1/destruction")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert r.text == isodatetime(destruction_time)

    r = await client.get("/test/jobs/1/executionduration")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert r.text == "600"

    r = await client.get("/test/jobs/1/owner")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert r.text == test_username

    r = await client.get("/test/jobs/1/parameters")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    expected = SimpleXmlParameters.from_xml(JOB_PARAMETERS)
    assert SimpleXmlParameters.from_xml(r.text) == expected

    r = await client.get("/test/jobs/1/phase")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert r.text == "PENDING"

    r = await client.get("/test/jobs/1/quote")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert r.text == ""

    # Modify various settings. Validators will be tested elsewhere.
    now = current_datetime()
    r = await client.post(
        "/test/jobs/1/destruction", data={"DESTRUCTION": isodatetime(now)}
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"

    r = await client.post(
        "/test/jobs/1/executionduration", data={"ExecutionDuration": 300}
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"

    # Retrieve the modified job and check that the new values are recorded.
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "1",
            "PENDING",
            isodatetime(job.creation_time),
            "300",
            isodatetime(now),
        ),
    )

    # Delete the job.
    r = await client.delete("/test/jobs/1")
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs"
    r = await client.get("/test/jobs/1")
    assert r.status_code == 404

    # Create a new job and then delete it via POST.
    r = await client.post(
        "/test/jobs", data={"name": "Jane", "RUNID": "some-run-id"}
    )
    assert r.status_code == 303
    job = await job_service.get(test_token, "2")
    r = await client.get("/test/jobs/2")
    assert r.status_code == 200
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "2",
            "PENDING",
            isodatetime(job.creation_time),
            "600",
            isodatetime(job.destruction_time),
        ),
    )
    r = await client.post("/test/jobs/2", data={"ACTION": "DELETE"})
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs"
    r = await client.get("/test/jobs/2")
    assert r.status_code == 404


@pytest.mark.asyncio
async def test_redirects(
    app: FastAPI, test_token: str, uws_factory: UWSFactory
) -> None:
    """Test the scheme in the redirect URLs.

    When running in a Kubernetes cluster behind an ingress that terminates
    TLS, the request as seen by the application will be ``http``, but we want
    the redirect URLs to honor ``X-Forwarded-Proto`` and thus use ``https``.
    We also want to honor the ``Host`` header.
    """
    job_service = uws_factory.create_job_service()
    await job_service.create(
        test_token, SimpleParameters(name="Peter"), run_id="some-run-id"
    )

    # Try various actions that result in redirects and ensure the redirect is
    # correct.
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://foo.com/",
        headers={"X-Auth-Request-Token": test_token},
    ) as client:
        r = await client.post(
            "/test/jobs/1/destruction",
            headers={
                "Host": "example.com",
                "X-Forwarded-For": "10.10.10.10",
                "X-Forwarded-Proto": "https",
                "X-Forwarded-Host": "foo.com",
            },
            data={"DESTRUCTION": "2021-09-10T10:01:02Z"},
        )
        assert r.status_code == 303
        assert r.headers["Location"] == "https://example.com/test/jobs/1"

        r = await client.post(
            "/test/jobs/1/executionduration",
            headers={
                "Host": "example.com",
                "X-Forwarded-For": "10.10.10.10",
                "X-Forwarded-Proto": "https",
                "X-Forwarded-Host": "foo.com",
            },
            data={"ExecutionDuration": 300},
        )
        assert r.status_code == 303
        assert r.headers["Location"] == "https://example.com/test/jobs/1"

        r = await client.delete(
            "/test/jobs/1",
            headers={
                "Host": "example.com",
                "X-Forwarded-For": "10.10.10.10",
                "X-Forwarded-Proto": "https",
                "X-Forwarded-Host": "foo.com",
            },
        )
        assert r.status_code == 303
        assert r.headers["Location"] == "https://example.com/test/jobs"


@pytest.mark.asyncio
async def test_presigned_url(
    client: AsyncClient,
    test_token: str,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    uws_config: UWSConfig,
) -> None:
    r = await client.post(
        "/test/jobs?phase=RUN", data={"runid": "some-run-id", "name": "Jane"}
    )
    assert r.status_code == 303
    await runner.mark_in_progress(test_token, "1")

    # Tell the queue the job is finished, with an https URL.
    results = [
        WorkerResult(
            result_id="cutout",
            url="https://example.com/some/path",
            mime_type="application/fits",
        )
    ]
    job = await runner.mark_complete(test_token, "1", results)

    # Check the job results, which should pass that URL through unchanged.
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert job.start_time
    assert job.end_time
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        FINISHED_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            "600",
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )


def validate_destruction(destruction: datetime, job: Job) -> datetime:
    max_destruction = current_datetime() + timedelta(days=1)
    return min(destruction, max_destruction)


def validate_execution_duration(
    duration: timedelta | None, job: Job
) -> timedelta | None:
    max_duration = timedelta(seconds=200)
    if not duration:
        return max_duration
    else:
        return min(duration, max_duration)


@pytest.mark.asyncio
async def test_validators(
    client: AsyncClient,
    test_token: str,
    arq_queue: MockArqQueue,
    uws_factory: UWSFactory,
    uws_config: UWSConfig,
) -> None:
    uws_config.validate_destruction = validate_destruction
    uws_config.validate_execution_duration = validate_execution_duration
    job_service = uws_factory.create_job_service()
    await job_service.create(test_token, SimpleParameters(name="Tiffany"))

    # Change the destruction time, first to something that should be honored
    # and then something that should be overridden.
    destruction = current_datetime() + timedelta(hours=1)
    r = await client.post(
        "/test/jobs/1/destruction",
        data={"desTRUcTiON": isodatetime(destruction)},
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"
    r = await client.get("/test/jobs/1/destruction")
    assert r.status_code == 200
    assert r.text == isodatetime(destruction)
    destruction = current_datetime() + timedelta(days=5)
    expected = current_datetime() + timedelta(days=1)
    r = await client.post(
        "/test/jobs/1/destruction",
        data={"destruction": isodatetime(destruction)},
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"
    r = await client.get("/test/jobs/1/destruction")
    assert r.status_code == 200
    seen = datetime.fromisoformat(r.text)
    assert seen >= expected - timedelta(seconds=5)
    assert seen <= expected + timedelta(seconds=5)

    # Now do the same thing for execution duration.
    r = await client.post(
        "/test/jobs/1/executionduration", data={"exECUTionduRATION": 100}
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"
    r = await client.get("/test/jobs/1/executionduration")
    assert r.status_code == 200
    assert r.text == "100"
    r = await client.post(
        "/test/jobs/1/executionduration", data={"exECUTionduRATION": 250}
    )
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/1"
    r = await client.get("/test/jobs/1/executionduration")
    assert r.status_code == 200
    assert r.text == "200"



"""Additional types for Pydantic models."""

from __future__ import annotations

import os
from datetime import datetime, timedelta
from typing import Annotated

from pydantic import (
    AfterValidator,
    BeforeValidator,
    PlainSerializer,
    UrlConstraints,
)
from pydantic_core import Url

from safir.datetime import isodatetime, parse_timedelta

from ._validators import normalize_datetime, normalize_isodatetime

__all__ = [
    "EnvAsyncPostgresDsn",
    "EnvRedisDsn",
    "HumanTimedelta",
    "IvoaIsoDatetime",
    "SecondsTimedelta",
    "UtcDatetime",
]


def _validate_env_async_postgres_dsn(v: Url) -> Url:
    """Possibly adjust a PostgreSQL DSN based on environment variables.

    When run via tox and tox-docker, the PostgreSQL hostname and port will be
    randomly selected and exposed only in environment variables. We have to
    patch that into the database URL at runtime since `tox doesn't have a way
    of substituting it into the environment
    <https://github.com/tox-dev/tox-docker/issues/55>`__.
    """
    if port := os.getenv("POSTGRES_5432_TCP_PORT"):
        return Url.build(
            scheme=v.scheme,
            username=v.username,
            password=v.password,
            host=os.getenv("POSTGRES_HOST", v.unicode_host() or "localhost"),
            port=int(port),
            path=v.path.lstrip("/") if v.path else v.path,
            query=v.query,
            fragment=v.fragment,
        )
    else:
        return v


type EnvAsyncPostgresDsn = Annotated[
    Url,
    UrlConstraints(
        host_required=True,
        allowed_schemes=["postgresql", "postgresql+asyncpg"],
    ),
    AfterValidator(_validate_env_async_postgres_dsn),
]
"""Async PostgreSQL data source URL honoring Docker environment variables.

Unlike the standard Pydantic ``PostgresDsn`` type, this type does not support
multiple hostnames because Safir's database library does not support multiple
hostnames.
"""


def _validate_env_redis_dsn(v: Url) -> Url:
    """Possibly adjust a Redis DSN based on environment variables.

    When run via tox and tox-docker, the Redis hostname and port will be
    randomly selected and exposed only in environment variables. We have to
    patch that into the Redis URL at runtime since `tox doesn't have a way of
    substituting it into the environment
    <https://github.com/tox-dev/tox-docker/issues/55>`__.
    """
    if port := os.getenv("REDIS_6379_TCP_PORT"):
        return Url.build(
            scheme=v.scheme,
            username=v.username,
            password=v.password,
            host=os.getenv("REDIS_HOST", v.unicode_host() or "localhost"),
            port=int(port),
            path=v.path.lstrip("/") if v.path else v.path,
            query=v.query,
            fragment=v.fragment,
        )
    else:
        return v


type EnvRedisDsn = Annotated[
    Url,
    UrlConstraints(
        allowed_schemes=["redis"],
        default_host="localhost",
        default_port=6379,
        default_path="/0",
    ),
    AfterValidator(_validate_env_redis_dsn),
]
"""Redis data source URL honoring Docker environment variables.

Unlike the standard Pydantic ``RedisDsn`` type, this does not support the
``rediss`` scheme, which indicates the use of TLS.
"""


def _validate_human_timedelta(v: str | float | timedelta) -> float | timedelta:
    if not isinstance(v, str):
        return v
    try:
        return float(v)
    except ValueError:
        return parse_timedelta(v)


type HumanTimedelta = Annotated[
    timedelta,
    BeforeValidator(_validate_human_timedelta),
    PlainSerializer(
        lambda t: t.total_seconds(), return_type=float, when_used="json"
    ),
]
"""Parse a human-readable string into a `datetime.timedelta`.

Accepts as input an integer or float (or stringified integer or float) number
of seconds, an already-parsed `~datetime.timedelta`, or a string consisting of
one or more sequences of numbers and duration abbreviations, separated by
optional whitespace.  Whitespace at the beginning and end of the string is
ignored. The supported abbreviations are:

- Week: ``weeks``, ``week``, ``w``
- Day: ``days``, ``day``, ``d``
- Hour: ``hours``, ``hour``, ``hr``, ``h``
- Minute: ``minutes``, ``minute``, ``mins``, ``min``, ``m``
- Second: ``seconds``, ``second``, ``secs``, ``sec``, ``s``

If several are present, they must be given in the above order. Example
valid strings are ``8d`` (8 days), ``4h 3minutes`` (four hours and three
minutes), and ``5w4d`` (five weeks and four days).
"""

type SecondsTimedelta = Annotated[
    timedelta,
    BeforeValidator(lambda v: v if not isinstance(v, str) else float(v)),
    PlainSerializer(
        lambda t: t.total_seconds(), return_type=float, when_used="json"
    ),
]
"""Parse a float number of seconds into a `datetime.timedelta`.

Accepts as input an integer or float (or stringified integer or float) number
of seconds or an already-parsed `~datetime.timedelta`. Compared to the
built-in Pydantic handling of `~datetime.timedelta`, an integer number of
seconds as a string is accepted, and ISO 8601 durations are not supported.
"""

type UtcDatetime = Annotated[datetime, AfterValidator(normalize_datetime)]
"""Coerce a `~datetime.datetime` to UTC.

Accepts as input all of the normal Pydantic representations of a
`~datetime.datetime`, but then forces the result to be timezone-aware and in
UTC.
"""

type IvoaIsoDatetime = Annotated[
    datetime,
    BeforeValidator(normalize_isodatetime),
    PlainSerializer(isodatetime, return_type=str, when_used="json"),
]
"""Accept the ISO datetime format required by IVOA standards.

The IVOA DALI standard requires timestamps to be formatted as in ISO 8601 but
without full timezone information. Either the timezone should be omitted or
the time portion should end in ``Z``, and in either case the timestamp should
be interpreted as UTC.

This type accepts the input formats that DALI accepts, provides a
timezone-aware `~datetime.datetime` in UTC, and always serializes to the ISO
8601 format ending in ``Z``.
"""



"""Middleware for IVOA services."""

from copy import copy
from urllib.parse import parse_qsl, urlencode

from starlette.types import ASGIApp, Receive, Scope, Send

__all__ = [
    "CaseInsensitiveFormMiddleware",
    "CaseInsensitiveQueryMiddleware",
]


class CaseInsensitiveQueryMiddleware:
    """Make query parameter keys all lowercase.

    Unfortunately, several IVOA standards require that query parameters be
    case-insensitive, which is not supported by modern HTTP web frameworks.
    This middleware attempts to work around this by lowercasing the query
    parameter keys before the request is processed, allowing normal FastAPI
    query parsing to then work without regard for case. This, in turn, permits
    FastAPI to perform input validation on GET parameters, which would
    otherwise only happen if the case used in the request happened to match
    the case used in the function signature.

    This unfortunately doesn't handle POST, so routes that accept POST will
    need to parse the POST data case-insensitively in the handler or a
    dependency.

    Based on `fastapi#826 <https://github.com/fastapi/fastapi/issues/826>`__.
    """

    def __init__(self, app: ASGIApp) -> None:
        self._app = app

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        if scope["type"] != "http" or not scope.get("query_string"):
            await self._app(scope, receive, send)
            return
        scope = copy(scope)
        params = [(k.lower(), v) for k, v in parse_qsl(scope["query_string"])]
        scope["query_string"] = urlencode(params).encode()
        await self._app(scope, receive, send)
        return


class CaseInsensitiveFormMiddleware:
    """Make POST parameter keys all lowercase.

    This middleware attempts to work around case-sensitivity issues by
    lowercasing POST parameter keys before the request is processed. This
    allows normal FastAPI parsing to work without regard for case, permitting
    FastAPI to perform input validation on the POST parameters.
    """

    def __init__(self, app: ASGIApp) -> None:
        """Initialize the middleware with the ASGI application.

        Parameters
        ----------
        app
            The ASGI application to wrap.
        """
        self._app = app

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        """Process request set query parameters and POST body keys to
        lowercase.
        """
        if scope["type"] != "http":
            await self._app(scope, receive, send)
            return

        scope = copy(scope)

        if scope["method"] == "POST" and self._is_form_data(scope):
            receive = self._wrapped_receive(receive)

        await self._app(scope, receive, send)

    @staticmethod
    def _is_form_data(scope: Scope) -> bool:
        """Check if the request contains form data.

        Parameters
        ----------
        scope
            The request scope.

        Returns
        -------
        bool
            True if the request contains form data, False otherwise.
        """
        headers = {
            k.decode("latin-1"): v.decode("latin-1")
            for k, v in scope.get("headers", [])
        }
        content_type = headers.get("content-type", "")
        return content_type.startswith("application/x-www-form-urlencoded")

    @staticmethod
    async def _get_body(receive: Receive) -> bytes:
        """Read the entire request body.

        Parameters
        ----------
        receive
            The receive function to read messages from.

        Returns
        -------
        bytes
            The entire request body.
        """
        body = b""
        more_body = True
        while more_body:
            message = await receive()
            body += message.get("body", b"")
            more_body = message.get("more_body", False)
        return body

    @staticmethod
    async def _process_form_data(body: bytes) -> bytes:
        """Process the body, lowercasing keys of form data.

        Parameters
        ----------
        body
            The request body.

        Returns
        -------
        bytes
            The processed request body with lowercased keys.
        """
        body_str = body.decode("utf-8")
        parsed = parse_qsl(body_str)
        lowercased = [(key.lower(), value) for key, value in parsed]
        processed = urlencode(lowercased)
        return processed.encode("utf-8")

    def _wrapped_receive(self, receive: Receive) -> Receive:
        """Wrap the receive function to process form data.

        Parameters
        ----------
        receive
            The receive function to wrap.

        Returns
        -------
        Receive
            The wrapped receive function.
        """
        processed = False

        async def inner() -> dict:
            nonlocal processed
            if processed:
                return {
                    "type": "http.request",
                    "body": b"",
                    "more_body": False,
                }

            body = await self._get_body(receive)
            processed_body = await self._process_form_data(body)
            processed = True
            return {
                "type": "http.request",
                "body": processed_body,
                "more_body": False,
            }

        return inner
